# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Text editor implementation"""


import os.path
import logging
from cdmpyparser import getBriefModuleInfoFromMemory
from ui.qt import (Qt, QTimer, pyqtSignal, QEvent, QToolTip,
                   QCursor, QApplication, QTextOption, QAction,
                   QPlainTextEdit)
from ui.mainwindowtabwidgetbase import MainWindowTabWidgetBase
from ui.completer import CodeCompleter
from ui.findinfiles import ItemToSearchIn, getSearchItemIndex
from ui.calltip import Calltip
from utils.globals import GlobalData
from utils.settings import Settings
from utils.encoding import (readEncodedFile, detectEolString,
                            detectWriteEncoding, writeEncodedFile)
from utils.fileutils import getFileProperties, isPythonMime, isMarkdownMime
from utils.diskvaluesrelay import setFileEncoding, getFileEncoding
from autocomplete.bufferutils import getContext
from autocomplete.completelists import (getCompletionList,
                                        getDefinitions, getOccurrences,
                                        getCallSignatures)
from .qpartwrap import QutepartWrapper
from .editorcontextmenus import EditorContextMenuMixin
from .linenomargin import CDMLineNumberMargin
from .flakesmargin import CDMFlakesMargin
from .bpmargin import CDMBreakpointMargin


CTRL_SHIFT = int(Qt.ShiftModifier | Qt.ControlModifier)
SHIFT = int(Qt.ShiftModifier)
CTRL = int(Qt.ControlModifier)
ALT = int(Qt.AltModifier)
CTRL_KEYPAD = int(Qt.KeypadModifier | Qt.ControlModifier)
NO_MODIFIER = int(Qt.NoModifier)


class TextEditor(QutepartWrapper, EditorContextMenuMixin):

    """Text editor implementation"""

    sigEscapePressed = pyqtSignal()
    sigCFlowSyncRequested = pyqtSignal(int, int, int)

    def __init__(self, parent, debugger):
        self._parent = parent
        QutepartWrapper.__init__(self, parent)
        EditorContextMenuMixin.__init__(self)

        self.setAttribute(Qt.WA_KeyCompression)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        skin = GlobalData().skin
        self.setPaper(skin['nolexerPaper'])
        self.setColor(skin['nolexerColor'])

        self.onTextZoomChanged()
        self.__initMargins(debugger)

        self.cursorPositionChanged.connect(self._onCursorPositionChanged)

        self.__skipChangeCursor = False

        self.__openedLine = None

        self.setFocusPolicy(Qt.StrongFocus)
        self.indentWidth = 4

        self.updateSettings()

        # Completion support
        self.__completionPrefix = ''
        self.__completionLine = -1
        self.__completionPos = -1
        self.__completer = CodeCompleter(self)
        self.__inCompletion = False
        self.__completer.activated.connect(self.insertCompletion)
        self.__lastTabPosition = None

        # Calltip support
        self.__calltip = None
        self.__callPosition = None
        self.__calltipTimer = QTimer(self)
        self.__calltipTimer.setSingleShot(True)
        self.__calltipTimer.timeout.connect(self.__onCalltipTimer)

        self.__initHotKeys()
        self.installEventFilter(self)

    def _drawIndentMarkersAndEdge(self, paintEventRect):
        """"Overriden function"""
        # At the termination time the indenter is set to None to release
        # the editor reference. However a paint event may still appear for
        # the editor to be closed. The paint event uses the indenter in the
        # qutepart code. So the condition below prevents a crash
        if self._indenter is not None:
            QutepartWrapper._drawIndentMarkersAndEdge(self, paintEventRect)

    def dedentLine(self):
        """Dedent the current line or selection"""
        self.decreaseIndentAction.activate(QAction.Trigger)

    def __initHotKeys(self):
        """Initializes a map for the hot keys event filter"""
        self.autoIndentLineAction.setShortcut('Ctrl+Shift+I')

        self.invokeCompletionAction.setEnabled(False)
        self.invokeCompletionAction.deleteLater()
        self.invokeCompletionAction = None

        self.__hotKeys = {
            CTRL_SHIFT: {Qt.Key_T: self.onJumpToTop,
                         Qt.Key_M: self.onJumpToMiddle,
                         Qt.Key_B: self.onJumpToBottom},
            SHIFT: {Qt.Key_Delete: self.onShiftDel,
                    Qt.Key_Backtab: self.dedentLine,
                    Qt.Key_End: self.onShiftEnd,
                    Qt.Key_Home: self.onShiftHome},
            CTRL: {Qt.Key_X: self.onShiftDel,
                   Qt.Key_C: self.onCtrlC,
                   Qt.Key_Insert: self.onCtrlC,
                   Qt.Key_Apostrophe: self.onHighlight,
                   Qt.Key_Period: self.onNextHighlight,
                   Qt.Key_Comma: self.onPrevHighlight,
                   Qt.Key_M: self.onCommentUncomment,
                   Qt.Key_Space: self.onAutoComplete,
                   Qt.Key_F1: self.onTagHelp,
                   Qt.Key_Backslash: self.onGotoDefinition,
                   Qt.Key_BracketRight: self.onOccurences,
                   Qt.Key_Slash: self.onShowCalltip,
                   Qt.Key_Minus: Settings().onTextZoomOut,
                   Qt.Key_Equal: Settings().onTextZoomIn,
                   Qt.Key_0: Settings().onTextZoomReset,
                   Qt.Key_Home: self.onFirstChar,
                   Qt.Key_End: self.onLastChar,
                   Qt.Key_B: self.highlightInOutline,
                   Qt.Key_QuoteLeft: self.highlightInCFlow},
            ALT: {Qt.Key_U: self.onScopeBegin},
            CTRL_KEYPAD: {Qt.Key_Minus: Settings().onTextZoomOut,
                          Qt.Key_Plus: Settings().onTextZoomIn,
                          Qt.Key_0: Settings().onTextZoomReset},
            NO_MODIFIER: {Qt.Key_Home: self.onHome,
                          Qt.Key_End: self.moveToLineEnd,
                          Qt.Key_F12: self.makeLineFirst}}

        # Not all the derived classes need certain tool functionality
        if hasattr(self._parent, "getType"):
            widgetType = self._parent.getType()
            if widgetType in [MainWindowTabWidgetBase.PlainTextEditor]:
                if hasattr(self._parent, "onOpenImport"):
                    self.__hotKeys[CTRL][Qt.Key_I] = self._parent.onOpenImport
        if hasattr(self._parent, "onNavigationBar"):
            self.__hotKeys[NO_MODIFIER][Qt.Key_F2] = \
                self._parent.onNavigationBar

    # Arguments: obj, event
    def eventFilter(self, _, event):
        """Event filter to catch shortcuts on UBUNTU"""
        if event.type() == QEvent.KeyPress:
            key = event.key()
            if self.isReadOnly():
                if key in [Qt.Key_Delete, Qt.Key_Backspace, Qt.Key_Backtab,
                           Qt.Key_X, Qt.Key_Tab, Qt.Key_Space, Qt.Key_Slash,
                           Qt.Key_Z, Qt.Key_Y]:
                    return True

            modifiers = int(event.modifiers())
            try:
                self.__hotKeys[modifiers][key]()
                return True
            except KeyError:
                return False
            except Exception as exc:
                logging.warning(str(exc))
        return False

    def wheelEvent(self, event):
        """Mouse wheel event"""
        if QApplication.keyboardModifiers() == Qt.ControlModifier:
            angleDelta = event.angleDelta()
            if not angleDelta.isNull():
                if angleDelta.y() > 0:
                    Settings().onTextZoomIn()
                else:
                    Settings().onTextZoomOut()
            event.accept()
        else:
            QutepartWrapper.wheelEvent(self, event)

    def focusInEvent(self, event):
        """Enable Shift+Tab when the focus is received"""
        if self._parent.shouldAcceptFocus():
            QutepartWrapper.focusInEvent(self, event)
        else:
            self._parent.setFocus()

    def focusOutEvent(self, event):
        """Disable Shift+Tab when the focus is lost"""
        self.__completer.hide()
        if not self.__inCompletion:
            self.__resetCalltip()
        QutepartWrapper.focusOutEvent(self, event)

    def updateSettings(self):
        """Updates the editor settings"""
        settings = Settings()

        if settings['verticalEdge']:
            self.lineLengthEdge = settings['editorEdge']
            self.lineLengthEdgeColor = GlobalData().skin['edgeColor']
            self.drawSolidEdge = True
        else:
            self.lineLengthEdge = None

        self.drawAnyWhitespace = settings['showSpaces']
        self.drawIncorrectIndentation = settings['showSpaces']

        if settings['currentLineVisible']:
            skin = GlobalData().skin
            self.currentLineColor = skin['currentLinePaper']
        else:
            self.currentLineColor = None

        if settings['lineWrap']:
            self.setWordWrapMode(QTextOption.WrapAnywhere)
        else:
            self.setWordWrapMode(QTextOption.NoWrap)

        if hasattr(self._parent, "getNavigationBar"):
            navBar = self._parent.getNavigationBar()
            if navBar:
                navBar.updateSettings()

        # Workaround: after changing the current line highlight visibility the
        # editor does not update it immediately
        self._updateExtraSelections()

    def __initMargins(self, debugger):
        """Initializes the editor margins"""
        self.addMargin(CDMLineNumberMargin(self))
        self.addMargin(CDMFlakesMargin(self))
        self.getMargin('cdm_flakes_margin').setVisible(False)

        if debugger:
            self.addMargin(CDMBreakpointMargin(self, debugger))
            self.getMargin('cdm_bpoint_margin').setVisible(False)

    def highlightCurrentDebuggerLine(self, line, asException):
        """Highlights the current debugger line"""
        margin = self.getMargin('cdm_flakes_margin')
        if margin:
            if asException:
                margin.setExceptionLine(line)
            else:
                margin.setCurrentDebugLine(line)

    def clearCurrentDebuggerLine(self):
        """Removes the current debugger line marker"""
        margin = self.getMargin('cdm_flakes_margin')
        if margin:
            margin.clearDebugMarks()

    def readFile(self, fileName):
        """Reads the text from a file"""
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        try:
            content, self.encoding = readEncodedFile(fileName)
            self.eol = detectEolString(content)

            # Copied from enki (enki/core/document.py: _readFile()):
            # Strip last EOL. Qutepart adds it when saving file
            if content.endswith('\r\n'):
                content = content[:-2]
            elif content.endswith('\n') or content.endswith('\r'):
                content = content[:-1]

            self.text = content

            self.mime, _, xmlSyntaxFile = getFileProperties(fileName)
            if xmlSyntaxFile:
                self.detectSyntax(xmlSyntaxFile)

            self.document().setModified(False)
        except:
            QApplication.restoreOverrideCursor()
            raise

        QApplication.restoreOverrideCursor()

    def writeFile(self, fileName):
        """Writes the text to a file"""
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

        if Settings()['removeTrailingOnSave']:
            self.removeTrailingWhitespaces()

        try:
            encoding = detectWriteEncoding(self, fileName)
            if encoding is None:
                QApplication.restoreOverrideCursor()
                logging.error('Could not detect write encoding for ' +
                              fileName)
                return False

            writeEncodedFile(fileName, self.textForSaving(), encoding)
        except Exception as exc:
            logging.error(str(exc))
            QApplication.restoreOverrideCursor()
            return False

        self.encoding = encoding
        if self.explicitUserEncoding:
            userEncoding = getFileEncoding(fileName)
            if userEncoding != self.explicitUserEncoding:
                setFileEncoding(fileName, self.explicitUserEncoding)
            self.explicitUserEncoding = None

        self._parent.updateModificationTime(fileName)
        self._parent.setReloadDialogShown(False)
        QApplication.restoreOverrideCursor()
        return True

    def setReadOnly(self, mode):
        """Overridden version"""
        QPlainTextEdit.setReadOnly(self, mode)
        if mode:
            # Otherwise the cursor is suppressed in the RO mode
            self.setTextInteractionFlags(self.textInteractionFlags() |
                                         Qt.TextSelectableByKeyboard)
        self.increaseIndentAction.setEnabled(not mode)
        self.decreaseIndentAction.setEnabled(not mode)
        self.autoIndentLineAction.setEnabled(not mode)
        self.indentWithSpaceAction.setEnabled(not mode)
        self.unIndentWithSpaceAction.setEnabled(not mode)
        self.undoAction.setEnabled(not mode)
        self.redoAction.setEnabled(not mode)
        self.moveLineUpAction.setEnabled(not mode)
        self.moveLineDownAction.setEnabled(not mode)
        self.deleteLineAction.setEnabled(not mode)
        self.pasteLineAction.setEnabled(not mode)
        self.cutLineAction.setEnabled(not mode)
        self.duplicateLineAction.setEnabled(not mode)

    def keyPressEvent(self, event):
        """Handles the key press events"""
        key = event.key()
        if self.isReadOnly():
            # Space scrolls
            # Ctrl+X/Shift+Del/Alt+D/Alt+X deletes something
            if key in [Qt.Key_Delete, Qt.Key_Backspace,
                       Qt.Key_Space, Qt.Key_X, Qt.Key_Tab,
                       Qt.Key_Z, Qt.Key_Y]:
                return

            # Qutepart has its own handler and lets to insert new lines when
            # ENTER is clicked, so use the QPlainTextEdit
            return QPlainTextEdit.keyPressEvent(self, event)

        self.__skipChangeCursor = True
        if self.__completer.isVisible():
            self.__skipChangeCursor = False
            if key == Qt.Key_Escape:
                self.__completer.hide()
                self.setFocus()
                return
            # There could be backspace or printed characters only
            QutepartWrapper.keyPressEvent(self, event)
            QApplication.processEvents()
            if key == Qt.Key_Backspace:
                if self.__completionPrefix == '':
                    self.__completer.hide()
                    self.setFocus()
                else:
                    self.__completionPrefix = self.__completionPrefix[:-1]
                    self.__completer.setPrefix(self.__completionPrefix)
            else:
                self.__completionPrefix += event.text()
                self.__completer.setPrefix(self.__completionPrefix)
                if self.__completer.completionCount() == 0:
                    self.__completer.hide()
                    self.setFocus()

        elif key in [Qt.Key_Enter, Qt.Key_Return]:
            QApplication.processEvents()
            line, _ = self.cursorPosition

            QutepartWrapper.keyPressEvent(self, event)
            QApplication.processEvents()

            if line == self.__openedLine:
                self.lines[line] = ''

            # If the new line has one or more spaces then it is a candidate for
            # automatic trimming
            line, pos = self.cursorPosition
            text = self.lines[line]
            self.__openedLine = None
            if pos > 0 and len(text.strip()) == 0:
                self.__openedLine = line

        elif key in [Qt.Key_Up, Qt.Key_PageUp,
                     Qt.Key_Down, Qt.Key_PageDown]:
            line, _ = self.cursorPosition
            lineToTrim = line if line == self.__openedLine else None

            QutepartWrapper.keyPressEvent(self, event)
            QApplication.processEvents()

            if lineToTrim is not None:
                line, _ = self.cursorPosition
                if line != lineToTrim:
                    # The cursor was really moved to another line
                    self.lines[lineToTrim] = ''
            self.__openedLine = None

        elif key == Qt.Key_Escape:
            self.__resetCalltip()
            self.sigEscapePressed.emit()
            event.accept()

        elif key == Qt.Key_Tab:
            if self.selectedText:
                QutepartWrapper.keyPressEvent(self, event)
                self.__lastTabPosition = None
            else:
                line, pos = self.cursorPosition
                currentPosition = self.absCursorPosition
                if pos != 0:
                    char = self.lines[line][pos - 1]
                    if char not in [' ', ':', '{', '}', '[', ']', ',',
                                    '<', '>', '+', '!', ')'] and \
                       currentPosition != self.__lastTabPosition:
                        self.__lastTabPosition = currentPosition
                        self.onAutoComplete()
                        event.accept()
                    else:
                        QutepartWrapper.keyPressEvent(self, event)
                        self.__lastTabPosition = currentPosition
                else:
                    QutepartWrapper.keyPressEvent(self, event)
                    self.__lastTabPosition = currentPosition

        elif key == Qt.Key_Z and \
            int(event.modifiers()) == (Qt.ControlModifier + Qt.ShiftModifier):
            event.accept()

        elif key == Qt.Key_ParenLeft:
            if Settings()['editorCalltips']:
                QutepartWrapper.keyPressEvent(self, event)
                self.onShowCalltip(False)
            else:
                QutepartWrapper.keyPressEvent(self, event)
        else:
            # Special keyboard keys are delivered as 0 values
            if key != 0:
                self.__openedLine = None
                QutepartWrapper.keyPressEvent(self, event)

        self.__skipChangeCursor = False

    def _onCursorPositionChanged(self):
        """Triggered when the cursor changed the position"""
        self.__lastTabPosition = None
        line, _ = self.cursorPosition

        if self.__calltip:
            if self.__calltipTimer.isActive():
                self.__calltipTimer.stop()
            if self.absCursorPosition < self.__callPosition:
                self.__resetCalltip()
            else:
                self.__calltipTimer.start(500)

        if not self.__skipChangeCursor:
            if line == self.__openedLine:
                self.__openedLine = None
                return

            if self.__openedLine is not None:
                self.__skipChangeCursor = True
                self.lines[self.__openedLine] = ''
                self.__skipChangeCursor = False
                self.__openedLine = None

    def getCurrentPosFont(self):
        """Provides the font of the current character"""
        if self.lexer_ is not None:
            font = self.lexer_.font(self.styleAt(self.currentPosition()))
        else:
            font = self.font()
        font.setPointSize(font.pointSize() + self.getZoom())
        return font

    def onCommentUncomment(self):
        """Triggered when Ctrl+M is received"""
        if self.isReadOnly() or not self.isPythonBuffer():
            return

        with self:
            # Detect what we need - comment or uncomment
            line, _ = self.cursorPosition
            txt = self.lines[line]
            nonSpaceIndex = self.firstNonSpaceIndex(txt)
            if self.isCommentLine(line):
                # need to uncomment
                if nonSpaceIndex == len(txt) - 1:
                    # Strip the only '#' character
                    stripCount = 1
                else:
                    # Strip up to two characters if the next char is a ' '
                    if txt[nonSpaceIndex + 1] == ' ':
                        stripCount = 2
                    else:
                        stripCount = 1
                newTxt = txt[:nonSpaceIndex] + txt[nonSpaceIndex +
                                                   stripCount:]
                if not newTxt.strip():
                    newTxt = ''
                self.lines[line] = newTxt
            else:
                # need to comment
                if nonSpaceIndex is None:
                    self.lines[line] = '# '
                else:
                    newTxt = '# '.join((txt[:nonSpaceIndex],
                                        txt[nonSpaceIndex:]))
                    self.lines[line] = newTxt

            # Jump to the beginning of the next line
            if line + 1 < len(self.lines):
                line += 1
            self.cursorPosition = line, 0
            self.ensureLineOnScreen(line)

    def onAutoComplete(self):
        """Triggered when ctrl+space or TAB is clicked"""
        if self.isReadOnly():
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self.__inCompletion = True
        self.__completionPrefix = self.getWordBeforeCursor()
        words = getCompletionList(self, self._parent.getFileName())

        QApplication.restoreOverrideCursor()

        if len(words) == 0:
            self.setFocus()
            self.__inCompletion = False
            return

        self.__completer.setWordsList(words, self.font())
        self.__completer.setPrefix(self.__completionPrefix)

        count = self.__completer.completionCount()
        if count == 0:
            self.setFocus()
            self.__inCompletion = False
            return

        # Make sure the line is visible
        line, _ = self.cursorPosition
        self.ensureLineOnScreen(line + 1)

        # Remove the selection as it could be interpreted not as expected
        if self.selectedText:
            self.clearSelection()

        if count == 1:
            self.insertCompletion(self.__completer.currentCompletion())
        else:
            cRectangle = self.cursorRect()
            cRectangle.setLeft(cRectangle.left() + self.viewport().x())
            cRectangle.setTop(cRectangle.top() + self.viewport().y() + 2)
            self.__completer.complete(cRectangle)
            # If something was selected then the next tab does not need to
            # bring the completion again. This is covered in the
            # insertCompletion() method. Here we reset the last tab position
            # preliminary because it is unknown if something will be inserted.
            self.__lastTabPosition = None
        self.__inCompletion = False

    def onTagHelp(self):
        """Provides help for an item if available"""
        if not self.isPythonBuffer():
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        definitions = getDefinitions(self, self._parent.getFileName())
        QApplication.restoreOverrideCursor()

        parts = []
        for definition in definitions:
            header = 'Type: ' + definition[3]
            if definition[5]:
                header += '\nModule: ' + definition[5]
            parts.append(header + '\n\n' + definition[4])

        if parts:
            QToolTip.showText(self.mapToGlobal(self.cursorRect().bottomLeft()),
                              '<pre>' + '\n\n'.join(parts) + '</pre>')
        else:
            GlobalData().mainWindow.showStatusBarMessage(
                "Definition is not found")

    def makeLineFirst(self):
        """Make the cursor line the first on the screen"""
        currentLine, _ = self.cursorPosition
        self.setFirstVisibleLine(currentLine)

    def onJumpToTop(self):
        """Jumps to the first position of the first visible line"""
        self.cursorPosition = self.firstVisibleLine(), 0

    def onJumpToMiddle(self):
        """Jumps to the first line pos in a middle of the editing area"""
        # Count the number of the visible line
        count = 0
        firstVisible = self.firstVisibleLine()
        lastVisible = self.lastVisibleLine()
        candidate = firstVisible
        while candidate <= lastVisible:
            if self.isLineVisible(candidate):
                count += 1
            candidate += 1

        shift = int(count / 2)
        jumpTo = firstVisible
        while shift > 0:
            if self.isLineVisible(jumpTo):
                shift -= 1
            jumpTo += 1
        self.cursorPosition = jumpTo, 0

    def onJumpToBottom(self):
        """Jumps to the first position of the last line"""
        currentFirstVisible = self.firstVisibleLine()
        self.cursorPosition = self.lastVisibleLine(), 0
        safeLastVisible = self.lastVisibleLine()

        while self.firstVisibleLine() != currentFirstVisible:
            # Here: a partially visible last line caused scrolling. So the
            # cursor needs to be set to the previous visible line
            self.cursorPosition = currentFirstVisible, 0
            safeLastVisible -= 1
            while not self.isLineVisible(safeLastVisible):
                safeLastVisible -= 1
            self.cursorPosition = safeLastVisible, 0

    def onGotoDefinition(self):
        """The user requested a jump to definition"""
        if not self.isPythonBuffer():
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        definitions = getDefinitions(self, self._parent.getFileName())
        QApplication.restoreOverrideCursor()

        if definitions:
            if len(definitions) == 1:
                GlobalData().mainWindow.openFile(
                    definitions[0][0], definitions[0][1],
                    definitions[0][2] + 1)
            else:
                if hasattr(self._parent, "importsBar"):
                    self._parent.importsBar.showDefinitions(definitions)
        else:
            GlobalData().mainWindow.showStatusBarMessage(
                "Definition is not found")

    def onScopeBegin(self):
        """The user requested jumping to the current scope begin"""
        if self.isPythonBuffer():
            info = getBriefModuleInfoFromMemory(self.text)
            context = getContext(self, info, True)
            if context.getScope() != context.GlobalScope:
                GlobalData().mainWindow.jumpToLine(context.getLastScopeLine())

    def onShowCalltip(self, showMessage=True):
        """The user requested show calltip"""
        if self.__calltip is not None:
            self.__resetCalltip()
            return
        if not self.isPythonBuffer():
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        signatures = getCallSignatures(self, self._parent.getFileName())
        QApplication.restoreOverrideCursor()

        if not signatures:
            if showMessage:
                GlobalData().mainWindow.showStatusBarMessage(
                    "No calltip found")
            return

        # For the time being let's take only the first signature...
        calltipParams = []
        for param in signatures[0].params:
            calltipParams.append(param.description[len(param.type) + 1:])
        calltip = signatures[0].name + '(' + ', '.join(calltipParams) + ')'

        self.__calltip = Calltip(self)
        self.__calltip.showCalltip(calltip, signatures[0].index)

        line = signatures[0].bracket_start[0]
        column = signatures[0].bracket_start[1]
        self.__callPosition = self.mapToAbsPosition(line - 1, column)

    def __resetCalltip(self):
        """Hides the calltip and resets how it was shown"""
        self.__calltipTimer.stop()
        if self.__calltip is not None:
            self.__calltip.hide()
            self.__calltip = None
        self.__callPosition = None

    def resizeCalltip(self):
        """Resizes the calltip if so"""
        if self.__calltip:
            self.__calltip.resize()

    def __onCalltipTimer(self):
        """Handles the calltip update timer"""
        if self.__calltip:
            if self.absCursorPosition < self.__callPosition:
                self.__resetCalltip()
                return

            QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
            signatures = getCallSignatures(self, self._parent.getFileName())
            QApplication.restoreOverrideCursor()

            if not signatures:
                self.__resetCalltip()
                return

            line = signatures[0].bracket_start[0]
            column = signatures[0].bracket_start[1]
            callPosition = self.mapToAbsPosition(line - 1, column)

            if callPosition != self.__callPosition:
                self.__resetCalltip()
            else:
                # It is still the same call, check the commas
                self.__calltip.highlightParameter(signatures[0].index)

    def onOccurences(self):
        """The user requested a list of occurences"""
        if not self.isPythonBuffer():
            return
        if self._parent.getType() == MainWindowTabWidgetBase.VCSAnnotateViewer:
            return
        if not os.path.isabs(self._parent.getFileName()):
            GlobalData().mainWindow.showStatusBarMessage(
                "Please save the buffer and try again")
            return

        fileName = self._parent.getFileName()
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        definitions = getOccurrences(self, fileName)
        QApplication.restoreOverrideCursor()

        if len(definitions) == 0:
            GlobalData().mainWindow.showStatusBarMessage('No occurences found')
            return

        # There are found items
        GlobalData().mainWindow.showStatusBarMessage('')
        result = []
        for definition in definitions:
            fName = definition.module_path
            if not fName:
                fName = fileName
            lineno = definition.line
            index = getSearchItemIndex(result, fName)
            if index < 0:
                widget = GlobalData().mainWindow.getWidgetForFileName(fName)
                if widget is None:
                    uuid = ""
                else:
                    uuid = widget.getUUID()
                newItem = ItemToSearchIn(fName, uuid)
                result.append(newItem)
                index = len(result) - 1
            result[index].addMatch(definition.name, lineno)

        GlobalData().mainWindow.displayFindInFiles('', result)

    def insertCompletion(self, text):
        """Triggered when a completion is selected"""
        if text:
            currentWord = self.getCurrentWord()
            line, pos = self.cursorPosition
            prefixLength = len(self.__completionPrefix)

            if text != currentWord and text != self.__completionPrefix:
                with self:
                    lineContent = self.lines[line]
                    leftPart = lineContent[0:pos - prefixLength]
                    rightPart = lineContent[pos:]
                    self.lines[line] = leftPart + text + rightPart

            newPos = pos + len(text) - prefixLength
            self.cursorPosition = line, newPos

            self.__completionPrefix = ''
            self.__completer.hide()

            # The next time there is nothing to insert for sure
            self.__lastTabPosition = self.absCursorPosition

    def insertLines(self, text, line):
        """Inserts the given text into new lines starting from 1-based line"""
        toInsert = text.splitlines()
        with self:
            if line > 0:
                line -= 1
            for item in toInsert:
                self.lines.insert(line, item)
                line += 1

    def hideCompleter(self):
        """Hides the completer if visible"""
        self.__completer.hide()

    def clearAnalysisMessages(self):
        """Clears all the analysis markers"""
        self.getMargin('cdm_flakes_margin').clearAnalysisMessages()

    def setAnalysisMessages(self, messages, ccMessages):
        """Shows up analysis messages"""
        self.getMargin('cdm_flakes_margin').setAnalysisMessages(messages,
                                                                ccMessages)

    def highlightInCFlow(self):
        """Triggered when highlight in the control flow is requested"""
        if self.isPythonBuffer():
            line, pos = self.cursorPosition
            absPos = self.absCursorPosition
            self.sigCFlowSyncRequested.emit(absPos, line + 1, pos + 1)

    def setDebugMode(self, debugOn, disableEditing):
        """Called to switch between debug/development"""
        skin = GlobalData().skin
        if debugOn:
            if disableEditing:
                self.setLinenoMarginBackgroundColor(skin['marginPaperDebug'])
                self.setLinenoMarginForegroundColor(skin['marginColorDebug'])
                self.setReadOnly(True)
        else:
            self.setLinenoMarginBackgroundColor(skin['marginPaper'])
            self.setLinenoMarginForegroundColor(skin['marginColor'])
            self.setReadOnly(False)

        bpointMargin = self.getMargin('cdm_bpoint_margin')
        if bpointMargin:
            bpointMargin.setDebugMode(debugOn, disableEditing)

    def restoreBreakpoints(self):
        """Restores the breakpoints"""
        bpointMargin = self.getMargin('cdm_bpoint_margin')
        if bpointMargin:
            bpointMargin.restoreBreakpoints()

    def isLineBreakable(self):
        """True if a line is breakable"""
        bpointMargin = self.getMargin('cdm_bpoint_margin')
        if bpointMargin:
            return bpointMargin.isLineBreakable()
        return False

    def validateBreakpoints(self):
        """Checks breakpoints and deletes those which are invalid"""
        bpointMargin = self.getMargin('cdm_bpoint_margin')
        if bpointMargin:
            bpointMargin.validateBreakpoints()

    def isPythonBuffer(self):
        """True if it is a python buffer"""
        return isPythonMime(self.mime)

    def isMarkdownBuffer(self):
        """True if it is a markdown buffer"""
        return isMarkdownMime(self.mime)

    def setLinenoMarginBackgroundColor(self, color):
        """Sets the margins background"""
        linenoMargin = self.getMargin('cdm_line_number_margin')
        if linenoMargin:
            linenoMargin.setBackgroundColor(color)

    def setLinenoMarginForegroundColor(self, color):
        """Sets the lineno margin foreground color"""
        linenoMargin = self.getMargin('cdm_line_number_margin')
        if linenoMargin:
            linenoMargin.setForegroundColor(color)

    def terminate(self):
        """Overloaded version to pass the request to margins too"""
        self.__hotKeys = None
        self.cursorPositionChanged.disconnect(self._onCursorPositionChanged)

        self.__completer.activated.disconnect(self.insertCompletion)
        self.__completer.deleteLater()

        self.__calltipTimer.timeout.disconnect(self.__onCalltipTimer)
        self.__calltipTimer.deleteLater()

        self.terminateMenus()

        for margin in self.getMargins():
            margin.clear()
            if hasattr(margin, 'onClose'):
                margin.onClose()
            margin.deleteLater()
        self._margins = None
        self.disconnect()
        QutepartWrapper.terminate(self)

    def resizeEvent(self, event):
        """Resize the parent panels if required"""
        QutepartWrapper.resizeEvent(self, event)
        self.hideCompleter()
        if hasattr(self._parent, 'resizeBars'):
            self._parent.resizeBars()

    def getFileName(self):
        """Provides the currently edited file name"""
        return self._parent.getFileName()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""VCS plugin support: indicators for the status bar and project viewer"""

import os.path
from ui.qt import QColor, QPixmap, QFrame, QLabel, QPalette, QSize, Qt
from utils.pixmapcache import getPixmapLocation, getPixmap


MAX_TEXT_INDICATOR_LENGTH = 2
MAX_PIXMAP_INDICATOR_WIDTH = 16
MAX_PIXMAP_INDICATOR_HEIGHT = 16
BROKEN_INDICATOR = "??"


def buildColor(colorAsStr):
    """Converts saved color into QColor object"""
    colorAsStr = colorAsStr.strip()
    parts = colorAsStr.split(',')

    length = len(parts)
    if length == 3:
        return QColor(int(parts[0]), int(parts[1]),
                      int(parts[2]))
    if length == 4:
        return QColor(int(parts[0]), int(parts[1]),
                      int(parts[2]), int(parts[3]))
    raise Exception("Unexpected color format")


class VCSIndicator:

    """Holds an indicator properties"""

    def __init__(self, configLine):
        """Config line looks as follows:
           id:::pathOrString:::ForegroundColor:::BackgroundColor:::Tooltip
           It comes from a config file or from a plugin"""
        self.identifier = None
        self.pixmap = None
        self.text = None
        self.backgroundColor = None
        self.foregroundColor = None
        self.defaultTooltip = ""

        self.__parseConfigLine(configLine)

    def __parseConfigLine(self, configLine):
        """Fills the members"""
        self.__parseConfigTuple(configLine)
        self.__scalePixmap()

    def __setBrokenIndicator(self, msg):
        """Sets the indicator to the broken state"""
        self.text = BROKEN_INDICATOR
        self.backgroundColor = QColor(0, 255, 255)
        self.foregroundColor = QColor(255, 0, 0)
        self.defaultTooltip = msg

    def __parseConfigTuple(self, pluginIndicator):
        """Checks what plugin provided"""
        if len(pluginIndicator) != 5:
            raise Exception("Unexpected format of an indicator "
                            "description. Expected 5 values.")
        # ID field
        self.identifier = int(pluginIndicator[0])

        # Pixmap/text field
        try:
            if isinstance(pluginIndicator[1], QPixmap):
                self.pixmap = QPixmap(pluginIndicator[1])
            elif os.path.exists(pluginIndicator[1]):
                self.pixmap = QPixmap(pluginIndicator[1])
            elif getPixmapLocation(pluginIndicator[1]) is not None:
                self.pixmap = getPixmap(pluginIndicator[1])
            else:
                self.__setText(pluginIndicator[1])
        except:
            self.__setBrokenIndicator("Failed to get plugin indicator "
                                      "pixmap. Indicator id: " +
                                      str(self.identifier))
            return

        # Foreground color
        if pluginIndicator[2] is None:
            self.foregroundColor = None
        else:
            if type(pluginIndicator[2]) == str:
                self.foregroundColor = buildColor(pluginIndicator[2])
            else:
                self.foregroundColor = QColor(pluginIndicator[2])

        # Background color
        if pluginIndicator[3] is None:
            self.backgroundColor = None
        else:
            if type(pluginIndicator[3]) == str:
                self.backgroundColor = buildColor(pluginIndicator[3])
            else:
                self.backgroundColor = QColor(pluginIndicator[3])

        # Default tooltip
        if pluginIndicator[4] is None:
            self.defaultTooltip = ""
        else:
            self.defaultTooltip = str(pluginIndicator[4]).strip()

    def __setText(self, value):
        """Sets the indicator text"""
        if len(value) > MAX_TEXT_INDICATOR_LENGTH:
            self.text = value[:MAX_TEXT_INDICATOR_LENGTH]
        else:
            self.text = value
        self.text = self.text.strip()

    def __scalePixmap(self):
        """Scales the pixmap if necessary"""
        if self.pixmap is None:
            return

        if self.pixmap.width() > MAX_PIXMAP_INDICATOR_WIDTH or \
           self.pixmap.height() > MAX_PIXMAP_INDICATOR_HEIGHT:
            maxSize = QSize(MAX_PIXMAP_INDICATOR_WIDTH,
                            MAX_PIXMAP_INDICATOR_HEIGHT)
            self.pixmap = self.pixmap.scaled(maxSize, Qt.KeepAspectRatio)

    def isPixmap(self):
        """True if it is a pixmap label"""
        return self.pixmap is not None

    def draw(self, label):
        """Draws the indicator as members tell. label is QLabel"""
        label.setPalette(QLabel().palette())
        if self.isPixmap():
            label.setAutoFillBackground(False)
            label.setFrameStyle(QLabel().frameStyle())
            label.setPixmap(self.pixmap)
        else:
            label.setFrameStyle(QFrame.StyledPanel)
            label.setAutoFillBackground(True)
            palette = label.palette()
            if self.backgroundColor is not None:
                palette.setColor(QPalette.Background, self.backgroundColor)
            if self.foregroundColor is not None:
                palette.setColor(QPalette.Foreground, self.foregroundColor)
            label.setPalette(palette)
            label.setText(self.text)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""project properties dialog"""

import os
from os.path import relpath
import pwd
import socket
import datetime
import logging
from utils.project import getProjectProperties
from utils.misc import getLocaleDate
from utils.settings import SETTINGS_DIR
from utils.encoding import SUPPORTED_CODECS, isValidEncoding
from .qt import (Qt, QEvent, QObject, QDialog, QLineEdit, QGridLayout, QLabel,
                 QTextEdit, QDialogButtonBox, QVBoxLayout, QPushButton,
                 QFileDialog, QMessageBox, QListWidget, QAbstractItemView,
                 QApplication, QComboBox)
from .labels import FramedLabel
from .itemdelegates import NoOutlineHeightDelegate
from .completers import DirCompleter, FileCompleter


class ProjectPropertiesDialog(QDialog):

    """project properties dialog implementation"""

    def __init__(self, project=None, parent=None):
        QDialog.__init__(self, parent)

        # The dialog caller reads this member if the dialog was finished
        # successfully.
        self.absProjectFileName = None

        self.__createLayout()
        self.__project = project

        if project is None:
            # It a new project creation
            self.setWindowTitle("New Project Properties")

            userRecord = pwd.getpwuid(os.getuid())

            if not userRecord[5].endswith(os.path.sep):
                self.dirEdit.setText(userRecord[5] + os.path.sep)
            else:
                self.dirEdit.setText(userRecord[5])
            self.initialDirName = self.dirEdit.text()
            self.lastProjectName = ""

            if userRecord[4] != "":
                self.authorEdit.setText(userRecord[4].split(',')[0].strip())
            else:
                self.authorEdit.setText(userRecord[0])

            try:
                self.emailEdit.setText(userRecord[0] + "@" +
                                       socket.gethostname())
            except:
                pass

            self.versionEdit.setText("0.0.1")
            self.licenseEdit.setText("GPL v3")
            self.copyrightEdit.setText("Copyright (c) " +
                                       self.authorEdit.text() + ", " +
                                       str(datetime.date.today().year))
            self.creationDateEdit.setText(getLocaleDate())
            self.nameEdit.setFocus()

        elif isinstance(project, str):
            self.setWindowTitle("Viewing Project Properties")

            # This is viewing properties and the argument is the path to the
            # project file
            props = getProjectProperties(project)

            scriptName = props['scriptname']
            if not os.path.isabs(scriptName) and scriptName != "":
                scriptName = os.path.normpath(os.path.dirname(project) +
                                              os.path.sep + scriptName)
            mdDocName = props['mddocfile']
            if not os.path.isabs(mdDocName) and mdDocName != '':
                mdDocName = os.path.normpath(os.path.dirname(project) +
                                             os.path.sep + mdDocName)

            self.nameEdit.setText(os.path.basename(project))
            self.nameEdit.setToolTip("")
            self.dirEdit.setText(os.path.dirname(project))
            self.dirEdit.setToolTip("")
            self.scriptEdit.setText(scriptName)
            self.mdDocEdit.setText(mdDocName)
            self.versionEdit.setText(props['version'])
            self.authorEdit.setText(props['author'])
            self.emailEdit.setText(props['email'])
            self.licenseEdit.setText(props['license'])
            self.copyrightEdit.setText(props['copyright'])
            self.descriptionEdit.setText(props['description'])
            self.encodingCombo.setCurrentText(props['encoding'])
            self.creationDateEdit.setText(props['creationdate'])
            self.uuidEdit.setText(props['uuid'])
            self.uuidEdit.setToolTip(SETTINGS_DIR + props['uuid'] +
                                     os.path.sep +
                                     " (double click to copy path)")

            for item in props['importdirs']:
                self.importDirList.addItem(item)

            self.disableEditing()
        else:
            self.setWindowTitle("Editing Project Properties")

            # This is editing the loaded project.
            self.nameEdit.setText(os.path.basename(project.fileName))
            self.nameEdit.setToolTip("")
            self.dirEdit.setText(project.getProjectDir())
            self.dirEdit.setToolTip("")
            self.scriptEdit.setText(project.getProjectScript())
            self.mdDocEdit.setText(project.getStartupMarkdownFile())
            self.versionEdit.setText(project.props['version'])
            self.authorEdit.setText(project.props['author'])
            self.emailEdit.setText(project.props['email'])
            self.licenseEdit.setText(project.props['license'])
            self.copyrightEdit.setText(project.props['copyright'])
            self.descriptionEdit.setText(project.props['description'])
            self.encodingCombo.setCurrentText(project.props['encoding'])
            self.creationDateEdit.setText(project.props['creationdate'])
            self.uuidEdit.setText(project.props['uuid'])
            self.uuidEdit.setToolTip(project.userProjectDir +
                                     " (double click to copy path)")
            self.setReadOnly()

            for item in project.props['importdirs']:
                self.importDirList.addItem(item)
            if self.importDirList.count() > 0:
                self.importDirList.setCurrentRow(0)
                self.delImportDirButton.setEnabled(True)

            # The project could be the one belonging to another user
            # so there might be no write permissions.
            if not os.access(project.fileName, os.W_OK):
                # Disable editing
                self.setWindowTitle("Viewing Project Properties "
                                    "(no write permissions)")
                self.disableEditing()
            else:
                self.scriptEdit.setFocus()

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(600, 400)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)
        gridLayout = QGridLayout()

        # Project name
        nameLabel = QLabel("Project name:", self)
        gridLayout.addWidget(nameLabel, 0, 0, 1, 1)
        self.nameEdit = QLineEdit(self)
        self.nameEdit.setToolTip("Type a project name without a path")
        self.nameEdit.installEventFilter(self)
        gridLayout.addWidget(self.nameEdit, 0, 1, 1, 1)

        # Project dir
        dirLabel = QLabel("Project directory:", self)
        gridLayout.addWidget(dirLabel, 1, 0, 1, 1)
        self.dirEdit = QLineEdit(self)
        self.dirEdit.setToolTip("Not existed directories will be created")
        gridLayout.addWidget(self.dirEdit, 1, 1, 1, 1)
        self.dirButton = QPushButton(self)
        self.dirButton.setText("...")
        gridLayout.addWidget(self.dirButton, 1, 2, 1, 1)
        self.dirCompleter = DirCompleter(self.dirEdit)

        # Project script
        mainScriptLabel = QLabel("Main script:", self)
        gridLayout.addWidget(mainScriptLabel, 2, 0, 1, 1)
        self.scriptEdit = QLineEdit(self)
        self.scriptEdit.setToolTip("Project main script, "
                                   "used when the project is run")
        gridLayout.addWidget(self.scriptEdit, 2, 1, 1, 1)
        self.scriptButton = QPushButton("...", self)
        gridLayout.addWidget(self.scriptButton, 2, 2, 1, 1)
        self.fileCompleter = FileCompleter(self.scriptEdit)

        # Project markdown doc file
        mddocLabel = QLabel('Markdown doc file:', self)
        gridLayout.addWidget(mddocLabel, 3, 0, 1, 1)
        self.mdDocEdit = QLineEdit(self)
        self.mdDocEdit.setToolTip('Project documentation start file')
        gridLayout.addWidget(self.mdDocEdit, 3, 1, 1, 1)
        self.mdDocButton = QPushButton("...", self)
        gridLayout.addWidget(self.mdDocButton, 3, 2, 1, 1)
        self.mdFileCompleter = FileCompleter(self.mdDocEdit)

        # Import dirs
        importLabel = QLabel("Import directories:", self)
        importLabel.setAlignment(Qt.AlignTop)
        gridLayout.addWidget(importLabel, 4, 0, 1, 1)
        self.importDirList = QListWidget(self)
        self.importDirList.setAlternatingRowColors(True)
        self.importDirList.setSelectionMode(QAbstractItemView.SingleSelection)
        self.importDirList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.importDirList.setItemDelegate(NoOutlineHeightDelegate(4))
        self.importDirList.setToolTip("Directories where to look for "
                                      "project specific imports")
        gridLayout.addWidget(self.importDirList, 4, 1, 1, 1)

        self.addImportDirButton = QPushButton(self)
        self.addImportDirButton.setText("Add dir")
        self.delImportDirButton = QPushButton(self)
        self.delImportDirButton.setText("Delete dir")
        self.delImportDirButton.setEnabled(False)
        vLayout = QVBoxLayout()
        vLayout.addWidget(self.addImportDirButton)
        vLayout.addWidget(self.delImportDirButton)
        vLayout.addStretch(0)
        gridLayout.addLayout(vLayout, 4, 2, 1, 1)

        # Version
        versionLabel = QLabel("Version:", self)
        gridLayout.addWidget(versionLabel, 5, 0, 1, 1)
        self.versionEdit = QLineEdit(self)
        gridLayout.addWidget(self.versionEdit, 5, 1, 1, 1)

        # Author
        authorLabel = QLabel("Author:", self)
        gridLayout.addWidget(authorLabel, 6, 0, 1, 1)
        self.authorEdit = QLineEdit(self)
        gridLayout.addWidget(self.authorEdit, 6, 1, 1, 1)

        # E-mail
        emailLabel = QLabel("E-mail:", self)
        gridLayout.addWidget(emailLabel, 7, 0, 1, 1)
        self.emailEdit = QLineEdit(self)
        gridLayout.addWidget(self.emailEdit, 7, 1, 1, 1)

        # License
        licenseLabel = QLabel("License:", self)
        gridLayout.addWidget(licenseLabel, 8, 0, 1, 1)
        self.licenseEdit = QLineEdit(self)
        gridLayout.addWidget(self.licenseEdit, 8, 1, 1, 1)

        # Copyright
        copyrightLabel = QLabel("Copyright:", self)
        gridLayout.addWidget(copyrightLabel, 9, 0, 1, 1)
        self.copyrightEdit = QLineEdit(self)
        gridLayout.addWidget(self.copyrightEdit, 9, 1, 1, 1)

        # Description
        descriptionLabel = QLabel("Description:", self)
        descriptionLabel.setAlignment(Qt.AlignTop)
        gridLayout.addWidget(descriptionLabel, 10, 0, 1, 1)
        self.descriptionEdit = QTextEdit(self)
        self.descriptionEdit.setTabChangesFocus(True)
        self.descriptionEdit.setAcceptRichText(False)
        gridLayout.addWidget(self.descriptionEdit, 10, 1, 1, 1)

        # Default encoding
        encodingLabel = QLabel('Default encoding:', self)
        gridLayout.addWidget(encodingLabel, 11, 0, 1, 1)
        self.encodingCombo = QComboBox(self)
        self.encodingCombo.addItem('')
        self.encodingCombo.addItems(sorted(SUPPORTED_CODECS))
        self.encodingCombo.setEditable(True)
        gridLayout.addWidget(self.encodingCombo, 11, 1, 1, 1)

        # Creation date
        creationDateLabel = QLabel("Creation date:", self)
        gridLayout.addWidget(creationDateLabel, 12, 0, 1, 1)
        self.creationDateEdit = FramedLabel(parent=self)
        self.creationDateEdit.setToolTip("Double click to copy")
        gridLayout.addWidget(self.creationDateEdit, 12, 1, 1, 1)

        # Project UUID
        uuidLabel = QLabel("UUID:", self)
        gridLayout.addWidget(uuidLabel, 13, 0, 1, 1)
        self.uuidEdit = FramedLabel(text='',
                                    callback=self.__copyProjectPath,
                                    parent=self)
        gridLayout.addWidget(self.uuidEdit, 13, 1, 1, 1)

        verticalLayout.addLayout(gridLayout)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel |
                                     QDialogButtonBox.Ok)
        verticalLayout.addWidget(buttonBox)

        nameLabel.setBuddy(self.nameEdit)
        dirLabel.setBuddy(self.dirEdit)
        versionLabel.setBuddy(self.versionEdit)
        authorLabel.setBuddy(self.authorEdit)
        emailLabel.setBuddy(self.emailEdit)
        licenseLabel.setBuddy(self.licenseEdit)
        copyrightLabel.setBuddy(self.copyrightEdit)
        descriptionLabel.setBuddy(self.descriptionEdit)

        buttonBox.accepted.connect(self.onOKButton)
        buttonBox.rejected.connect(self.reject)
        self.dirButton.clicked.connect(self.onDirButton)
        self.scriptButton.clicked.connect(self.onScriptButton)
        self.mdDocButton.clicked.connect(self.onMdDocButton)
        self.importDirList.currentRowChanged.connect(self.onImportDirRowChanged)
        self.addImportDirButton.clicked.connect(self.onAddImportDir)
        self.delImportDirButton.clicked.connect(self.onDelImportDir)
        self.nameEdit.textEdited.connect(self.onProjectNameChanged)

        self.setTabOrder(self.nameEdit, self.dirEdit)
        self.setTabOrder(self.dirEdit, self.dirButton)
        self.setTabOrder(self.dirButton, self.scriptEdit)
        self.setTabOrder(self.scriptEdit, self.scriptButton)
        self.setTabOrder(self.scriptButton, self.mdDocEdit)
        self.setTabOrder(self.mdDocEdit, self.mdDocButton)
        self.setTabOrder(self.mdDocButton, self.importDirList)
        self.setTabOrder(self.importDirList, self.addImportDirButton)
        self.setTabOrder(self.addImportDirButton, self.delImportDirButton)
        self.setTabOrder(self.delImportDirButton, self.versionEdit)
        self.setTabOrder(self.versionEdit, self.authorEdit)
        self.setTabOrder(self.authorEdit, self.emailEdit)
        self.setTabOrder(self.emailEdit, self.licenseEdit)
        self.setTabOrder(self.licenseEdit, self.copyrightEdit)
        self.setTabOrder(self.copyrightEdit, self.descriptionEdit)
        self.setTabOrder(self.descriptionEdit, buttonBox)

    def eventFilter(self, obj, event):
        """Event filter for the project name field"""
        # Do not allow path separators
        if event.type() == QEvent.KeyPress:
            if event.key() == ord(os.path.sep):
                return True
        return QObject.eventFilter(self, obj, event)

    def onDirButton(self):
        """Displays a directory selection dialog"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog | QFileDialog.ShowDirsOnly
        dirName = QFileDialog.getExistingDirectory(
            self, "Select project directory", self.dirEdit.text(),
            options)
        if dirName:
            self.dirEdit.setText(os.path.normpath(dirName))

    def onScriptButton(self):
        """Displays a file selection dialog"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        scriptName = QFileDialog.getOpenFileName(
            self, 'Select project main script', self.dirEdit.text(),
            'Python Files (*.py);;All Files (*)',
            options=options)
        if isinstance(scriptName, tuple):
            scriptName = scriptName[0]
        if scriptName:
            self.scriptEdit.setText(os.path.normpath(scriptName))

    def onMdDocButton(self):
        """Displays an md file selection dialog"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        mdFileName = QFileDialog.getOpenFileName(
            self, 'Select project doc startup MD file', self.dirEdit.text(),
            'Markdown Files (*.md);;All Files (*)',
            options=options)
        if isinstance(mdFileName, tuple):
            mdFileName = mdFileName[0]
        if mdFileName:
            self.mdDocEdit.setText(os.path.normpath(mdFileName))

    def onImportDirRowChanged(self, row):
        """Triggered when a current row in the import dirs is changed"""
        self.delImportDirButton.setEnabled(row != -1)

    def onAddImportDir(self):
        """Displays a directory selection dialog"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog | QFileDialog.ShowDirsOnly
        dirName = QFileDialog.getExistingDirectory(
            self, 'Select import directory', self.dirEdit.text(),
            options=options)

        if not dirName:
            return

        # There are 2 cases: new project or
        # editing the existed project properties
        if self.__project is None:
            # It a new project; the project path could be editedd
            dirToInsert = dirName
        else:
            # This is an existed project; no way the project path is changed
            # Let's decide it a relative path should be used here
            if self.__project.isProjectDir(dirName):
                dirToInsert = relpath(dirName, self.dirEdit.text())
            else:
                dirToInsert = dirName

        index = 0
        while index < self.importDirList.count():
            if self.importDirList.item(index).text() == dirToInsert:
                logging.warning("The directory '%s' is already in the list of "
                                "imported directories and is not added.",
                                dirName)
                return
            index += 1

        self.importDirList.addItem(dirToInsert)
        self.importDirList.setCurrentRow(self.importDirList.count() - 1)

    def onDelImportDir(self):
        """Triggered when an import dir should be deleted"""
        rowToDelete = self.importDirList.currentRow()
        if  rowToDelete == -1:
            self.delImportDirButton.setEnabled(False)
            return

        self.importDirList.takeItem(rowToDelete)
        if self.importDirList.count() == 0:
            self.delImportDirButton.setEnabled(False)
        else:
            self.importDirList.setCurrentRow(self.importDirList.count() - 1)

    def onOKButton(self):
        """Checks that the mandatory fields are filled properly"""
        # The checks must be done for a new project only
        if not self.nameEdit.isEnabled():
            if self.__checkEncoding():
                self.accept()
            return

        # Check that the project name does not have path separators and is not
        # empty
        if not self.nameEdit.text().strip():
            QMessageBox.critical(self, "Error",
                                 "The project name must not be empty")
            return
        if os.path.sep in self.nameEdit.text():
            QMessageBox.critical(self, "Error",
                                 "The project name must not "
                                 "contain path separators")
            return

        # Check that the project directory is given
        dirName = self.dirEdit.text().strip()
        if not dirName:
            QMessageBox.critical(self, "Error",
                                 "The project directory must not be empty")
            return

        dirName = os.path.abspath(dirName)
        self.dirEdit.setText(dirName)
        # Check that the project file does not exist
        projectFileName = dirName
        if not projectFileName.endswith(os.path.sep):
            projectFileName += os.path.sep
        projectFileName += self.nameEdit.text().strip()
        if not projectFileName.endswith(".cdm3"):
            projectFileName += ".cdm3"

        if os.path.exists(projectFileName):
            QMessageBox.critical(self, "Error",
                                 "The project file " + projectFileName +
                                 " exists. Please provide another "
                                 "directory / project name.")
            return

        # Check that the project dir is not a file
        if os.path.exists(dirName):
            # It might be a link, so read it first
            dirName = os.path.realpath(dirName)
            if not os.path.exists(dirName):
                QMessageBox.critical(self, "Error",
                                     "Broken link: " + dirName)
                return
            if not os.path.isdir(dirName):
                QMessageBox.critical(self, "Error",
                                     "The project directory may not be a file")
                return
            # Check that the dir is writable
            if not os.access(dirName, os.W_OK):
                QMessageBox.critical(self, "Error",
                                     "You don't have write permissions on " +
                                     dirName)
                return
        else:
            # Create the directory
            try:
                os.makedirs(dirName)
            except OSError:
                QMessageBox.critical(self, "Error",
                                     "Cannot create the project directory")
                return

        if not self.__checkEncoding():
            return

        # Save the absolute file name for further reading it by the caller
        self.absProjectFileName = projectFileName

        # The minimum is provided so we can accept it
        self.accept()

    def __checkEncoding(self):
        """True if OK"""
        enc = self.encodingCombo.currentText().strip()
        if enc:
            if not isValidEncoding(enc):
                QMessageBox.critical(self, "Error",
                                     "Unsupported default project encoding")
                return False
        return True

    def onProjectNameChanged(self, newName):
        """Called when the project name changed"""
        if newName.endswith(".cdm3"):
            newName = newName[:-5]
        if self.dirEdit.text().strip() == (self.initialDirName +
                                           self.lastProjectName):
            self.dirEdit.setText(self.initialDirName + newName)
            self.lastProjectName = newName

    def setReadOnly(self):
        """Disables editing some fields"""
        self.dirEdit.setReadOnly(True)
        self.dirEdit.setFocusPolicy(Qt.NoFocus)
        self.dirEdit.setDisabled(True)
        self.dirButton.setDisabled(True)
        self.dirButton.setFocusPolicy(Qt.NoFocus)
        self.nameEdit.setReadOnly(True)
        self.nameEdit.setFocusPolicy(Qt.NoFocus)
        self.nameEdit.setDisabled(True)

    def disableEditing(self):
        """Disables all the editing"""
        self.nameEdit.setDisabled(True)
        self.dirEdit.setDisabled(True)
        self.dirButton.setDisabled(True)
        self.scriptEdit.setDisabled(True)
        self.scriptButton.setDisabled(True)
        self.mdDocEdit.setDisabled(True)
        self.mdDocButton.setDisabled(True)
        self.importDirList.setDisabled(True)
        self.addImportDirButton.setDisabled(True)
        self.delImportDirButton.setDisabled(True)
        self.versionEdit.setDisabled(True)
        self.authorEdit.setDisabled(True)
        self.emailEdit.setDisabled(True)
        self.licenseEdit.setDisabled(True)
        self.copyrightEdit.setDisabled(True)
        self.descriptionEdit.setDisabled(True)
        self.encodingCombo.setDisabled(True)

    def __copyProjectPath(self):
        """Copies the project path when a label is double clicked"""
        text = self.uuidEdit.text().strip()
        if text:
            path = SETTINGS_DIR + text + os.path.sep
            QApplication.clipboard().setText(path)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Debugger watchpoint"""


class Watchpoint:

    """Represents a single watchpoint"""

    def __init__(self, condition=None, special=False,
                 temporary=False, enabled=True, ignoreCount=0):
        self.__condition = condition
        self.__special = special
        self.__temporary = temporary
        self.__enabled = enabled
        self.__ignoreCount = 0

    def isValid(self):
        """True if the watchpoint is valid"""
        return self.__condition is not None

    def getSpecial(self):
        """Provides the special"""
        return self.__special

    def getCondition(self):
        """Provides the condition"""
        return self.__condition

    def isTemporary(self):
        """True if temporary"""
        return self.__temporary

    def isEnabled(self):
        """True if enabled"""
        return self.__enabled

    def getIgnoreCount(self):
        """Provides the ignore count"""
        return self.__ignoreCount

    def serialize(self):
        """Serializes the watchpoint to a string"""
        return {'condition': self.__condition,
                'special': self.__special,
                'temp': self.__temporary,
                'enabled': self.__enabled,
                'ignorecnt':self.__ignoreCount}

    def deserialize(self, source):
        """Deserializes the watchpoint"""
        self.__condition = source.get('condition', None)
        self.__special = source.get('special', False)
        self.__temporary = source.get('temp', False)
        self.__enabled = source.get('enabled', False)
        self.__ignoreCount = source.get('ignorecnt', 0)
        return self.isValid()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Annotated VCS viewer implementation"""


import os.path
from ui.qt import (Qt, QSize, QPoint, pyqtSignal, QToolBar, QFont,
                   QFontMetrics, QHBoxLayout, QWidget, QAction, QSizePolicy,
                   QToolTip)
from ui.mainwindowtabwidgetbase import MainWindowTabWidgetBase
from utils.fileutils import getFileProperties, isPythonMime
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from utils.importutils import getImportsList, resolveImports
from ui.importlist import ImportListWidget
from .texteditor import TextEditor


class VCSAnnotateViewer(TextEditor):

    """VCS blame view"""

    LINENUM_MARGIN = 0      # Matches the text editor
    REVISION_MARGIN = 1     # Introduced here
    FOLDING_MARGIN = 2      # Matches the text editor
    MESSAGES_MARGIN = 3     # Matches the text editor

    def __init__(self, parent):
        self.__maxLength = None

        TextEditor.__init__(self, parent, None)
        self.__revisionTooltipShown = False
        self.__initAlterRevisionMarker()
        self._updateDwellingTime()

    def __initAlterRevisionMarker(self):
        skin = GlobalData().skin
        self.__alterMarker = self.markerDefine(QsciScintilla.Background)
        self.setMarkerBackgroundColor(skin.revisionAlterPaper,
                                      self.__alterMarker)

    def setAnnotatedContent(self, shortName, text, lineRevisions, revisionInfo):
        """Sets the content"""
        self.__lineRevisions = lineRevisions
        self.__revisionInfo = revisionInfo

        fileType = self._parent.getFileType()
        if fileType in [DesignerFileType, LinguistFileType]:
            # special treatment for Qt-Linguist and Qt-Designer files
            self.encoding = 'latin-1'
        else:
            text, self.encoding = decode(text)

        self.detectEolString(text)
        self.setText(text)
        self.bindLexer(shortName, fileType)

        self.setModified(False)
        self.setReadOnly(True)

        self.__initAnnotateMargins()
        self.__setRevisionText()

    def __setRevisionText(self):
        """Sets the revision margin text"""
        for revNumber in self.__revisionInfo:
            author = self.__revisionInfo[revNumber]['author']
            if '@' in author:
                # Most probably this is an e-mail address. Leave just name.
                self.__revisionInfo[revNumber]['shortAuthor'] = author.split('@')[0]
            else:
                self.__revisionInfo[revNumber]['shortAuthor'] = author

        skin = GlobalData().skin
        revisionMarginFont = QFont(skin.lineNumFont)
        revisionMarginFont.setItalic(True)
        style = QsciStyle(-1, "Revision margin style",
                          skin.revisionMarginColor,
                          skin.revisionMarginPaper, revisionMarginFont)

        lineNumber = 0
        self.__maxLength = -1

        # Altering line background support
        currentRevision = -1
        needMarker = True

        for lineRevision in self.__lineRevisions:
            if lineRevision in self.__revisionInfo:
                marginText = " " + ":".join([str(lineRevision),
                    self.__revisionInfo[lineRevision]['shortAuthor']])
            else:
                marginText = " " + str(lineRevision)
            textLength = len(marginText)
            if textLength > self.__maxLength:
                self.__maxLength = textLength
            self.setMarginText(lineNumber, marginText, style)

            # Set the line background if needed
            if lineRevision != currentRevision:
                currentRevision = lineRevision
                needMarker = not needMarker
            if needMarker:
                self.markerAdd(lineNumber, self.__alterMarker)

            lineNumber += 1

        self.setRevisionMarginWidth()

    def detectRevisionMarginWidth(self):
        """Caculates the margin width depending on
           the margin font and the current zoom"""
        skin = GlobalData().skin
        font = QFont(skin.lineNumFont)
        font.setPointSize(font.pointSize() + self.getZoom())
        fontMetrics = QFontMetrics(font, self)
        return fontMetrics.width('W' * self.__maxLength) + 3

    def setRevisionMarginWidth(self):
        """Called when zooming is done to keep the width wide enough"""
        if self.__maxLength:
            self.setMarginWidth(self.REVISION_MARGIN,
                                self.detectRevisionMarginWidth())
        else:
            self.setMarginWidth(self.REVISION_MARGIN, 0)

    def __initAnnotateMargins(self):
        """Initializes the editor margins"""
        self.setMarginType(self.REVISION_MARGIN, self.TextMargin)
        self.setMarginMarkerMask(self.REVISION_MARGIN, 0)

        # Together with overriding _marginClicked(...) this
        # prevents selecting a line when the margin is clicked.
        self.setMarginSensitivity(self.REVISION_MARGIN, True)

    def _marginClicked(self, margin, line, modifiers):
        """Handles a click on the margin"""
        return

    def __getRevisionMarginTooltip(self, lineNumber):
        """lineNumber is zero based"""
        revisionNumber = self.__lineRevisions[lineNumber]
        if not revisionNumber in self.__revisionInfo:
            return None

        tooltip = "Revision: " + \
            str(revisionNumber) + "\n" \
            "Author: " + \
            self.__revisionInfo[revisionNumber]['author'] + "\n" \
            "Date: " + \
            str(self.__revisionInfo[revisionNumber]['date'])
        comment = self.__revisionInfo[revisionNumber]['message']
        if comment:
            tooltip += "\nComment: " + comment
        return tooltip

    def _updateDwellingTime(self):
        """There is always something to show"""
        self.SendScintilla(self.SCI_SETMOUSEDWELLTIME, 250)

    def _onDwellStart(self, position, x, y):
        """Triggered when mouse started to dwell"""
        if not self.underMouse():
            return

        marginNumber = self._marginNumber(x)
        if marginNumber == self.REVISION_MARGIN:
            self.__showRevisionTooltip(position, x, y)
        else:
            TextEditor._onDwellStart(self, position, x, y)

    def __showRevisionTooltip(self, position, x, y):
        """Shows the tooltip"""
        # Calculate the line
        pos = self.SendScintilla(self.SCI_POSITIONFROMPOINT, x, y)
        line, posInLine = self.lineIndexFromPosition(pos)

        tooltip = self.__getRevisionMarginTooltip(line)
        if tooltip:
            QToolTip.showText(self.mapToGlobal(QPoint(x, y)), tooltip)
            self.__revisionTooltipShown = True

    def _onDwellEnd(self, position, x, y):
        """Triggered when mouse ended to dwell"""
        if self.__revisionTooltipShown:
            self.__revisionTooltipShown = False
            QToolTip.hideText()

    def setLineNumMarginWidth(self):
        """Sets the line number margin widht"""
        TextEditor.setLineNumMarginWidth(self)
        self.setRevisionMarginWidth()

    def _contextMenuAboutToShow(self):
        """Enables/disables the appropriate menu items"""
        self._menuHighlightInOutline.setEnabled(self.isPythonBuffer())

        self.encodingMenu.setEnabled(False)
        self.pylintAct.setEnabled(False)
        self.pymetricsAct.setEnabled(False)
        self.runAct.setEnabled(False)
        self.runParamAct.setEnabled(False)
        self.profileAct.setEnabled(False)
        self.profileParamAct.setEnabled(False)
        self.importsDgmAct.setEnabled(False)
        self.importsDgmParamAct.setEnabled(False)


class VCSAnnotateViewerTabWidget(QWidget, MainWindowTabWidgetBase):

    """VCS annotate viewer tab widget"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent):

        MainWindowTabWidgetBase.__init__(self)
        QWidget.__init__(self, parent)

        self.__viewer = VCSAnnotateViewer(self)
        self.__viewer.sigEscapePressed.connect(self.__onEsc)
        self.__fileName = ""
        self.__shortName = ""

        self.__createLayout()
        self.__viewer.zoomTo(Settings()['zoom'])

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def __createLayout(self):
        """Creates the toolbar and layout"""
        # Buttons
        printButton = QAction(getIcon('printer.png'), 'Print', self)
        printButton.triggered.connect(self.__onPrint)
        printButton.setEnabled(False)
        printButton.setVisible(False)

        printPreviewButton = QAction(
            getIcon('printpreview.png'), 'Print preview', self)
        printPreviewButton.triggered.connect(self.__onPrintPreview)
        printPreviewButton.setEnabled(False)
        printPreviewButton.setVisible(False)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # The toolbar
        toolbar = QToolBar(self)
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.RightToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)

        toolbar.addAction(printPreviewButton)
        toolbar.addAction(printButton)
        toolbar.addWidget(spacer)

        self.__importsBar = ImportListWidget(self.__viewer)
        self.__importsBar.hide()

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)
        hLayout.addWidget(self.__viewer)
        hLayout.addWidget(toolbar)

        self.setLayout(hLayout)

    def updateStatus(self):
        """Updates the toolbar buttons status"""
        pass

    def __onPrint(self):
        """Triggered when the print button is pressed"""
        pass

    def __onPrintPreview(self):
        """triggered when the print preview button is pressed"""
        pass

    def setFocus(self):
        """Overridden setFocus"""
        self.__viewer.setFocus()

    def onOpenImport(self):
        """Triggered when Ctrl+I is received"""
        if isPythonMime(self.__editor.mime):
            # Take all the file imports and resolve them
            fileImports = getImportsList(self.__editor.text)
            if not fileImports:
                GlobalData().mainWindow.showStatusBarMessage(
                    "There are no imports")
            else:
                self.__onImportList(self.__fileName, fileImports)

    def __onImportList(self, fileName, imports):
        """Works with a list of imports"""
        # It has already been checked that the file is a Python one
        resolvedList, errors = resolveImports(fileName, imports)
        del errors  # errors are OK here
        if resolvedList:
            # Display the import selection widget
            self.__importsBar.showResolvedImports(resolvedList)
        else:
            GlobalData().mainWindow.showStatusBarMessage(
                "Could not resolve any imports")

    def resizeEvent(self, event):
        """Resizes the import selection dialogue if necessary"""
        QWidget.resizeEvent(self, event)
        self.resizeBars()

    def resizeBars(self):
        """Resize the bars if they are shown"""
        if not self.__importsBar.isHidden():
            self.__importsBar.resize()
        self.__viewer.resizeCalltip()

    def onRunScript(self, action=False):
        return True
    def onRunScriptSettings(self):
        return True
    def onProfileScript(self, action=False):
        return True
    def onProfileScriptSettings(self):
        return True
    def onImportDgm(self, action=None):
        return True
    def onImportDgmTuned(self):
        return True
    def shouldAcceptFocus(self):
        return True

    def setAnnotatedContent(self, shortName, text,
                            lineRevisions, revisionInfo):
        """Sets the content"""
        self.setShortName(shortName)
        self.__viewer.setAnnotatedContent(shortName, text,
                                          lineRevisions, revisionInfo)

    def writeFile(self, fileName):
        """Writes the text to a file"""
        return self.__viewer.writeFile(fileName)

    def updateModificationTime(self, fileName):
        """No need to implement"""
        return

    # Mandatory interface part is below

    def getEditor(self):
        """Provides the editor widget"""
        return self.__viewer

    def isModified(self):
        """Tells if the file is modified"""
        return False

    def getRWMode(self):
        """Tells if the file is read only"""
        return "RO"

    def getMime(self):
        """Provides the file type"""
        return self.__viewer.mime

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.VCSAnnotateViewer

    def getLanguage(self):
        """Tells the content language"""
        lang = self.__viewer.language()
        if lang:
            return lang
        return self.__viewer.mime if self.__viewer.mime else 'n/a'

    def setFileName(self, name):
        """Sets the file name"""
        raise Exception("Setting a file name for "
                        "annotate results is not applicable")

    def getEol(self):
        """Tells the EOL style"""
        return self.__viewer.getEolIndicator()

    def getLine(self):
        """Tells the cursor line"""
        line, _ = self.__viewer.cursorPosition
        return line

    def getPos(self):
        """Tells the cursor column"""
        _, pos = self.__viewer.cursorPosition
        return pos

    def getEncoding(self):
        """Tells the content encoding"""
        return self.__viewer.encoding

    def setEncoding(self, newEncoding):
        """Sets the new editor encoding"""
        self.__viewer.setEncoding(newEncoding)

    def getShortName(self):
        """Tells the display name"""
        return self.__shortName

    def setShortName(self, name):
        """Sets the display name"""
        self.__shortName = name



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Base class for all codimension plugins"""

import logging
from yapsy.IPlugin import IPlugin
from ui.qt import QObject, pyqtSignal
from utils.settings import SETTINGS_DIR


class CDMPluginBase(IPlugin, QObject):

    """Base class for all codimension plugin categories"""

    pluginLogMessage = pyqtSignal(int, str)

    def __init__(self):
        IPlugin.__init__(self)
        QObject.__init__(self)

        self.ide = IDEAccess(self)

    def activate(self, ideSettings, ideGlobalData):
        """Activates the plugin.

        Also saves references to the IDE settings and global data
        """
        IPlugin.activate(self)
        self.ide.activate(ideSettings, ideGlobalData)

    def deactivate(self):
        """Deactivates the plugin.

        Also clears references to the IDE settings and global data
        """
        self.ide.deactivate()
        IPlugin.deactivate(self)

    def getConfigFunction(self):
        """A plugin may provide a function for configuring it.

        If a plugin does not require any config parameters then None
        should be returned.
        By default no configuring is required.
        """
        return None


class ViewAndToolbar():

    """Incapsulates access to a certain view widget inside a side panel"""

    def __init__(self, widget, toolbar):
        self.widget = widget
        self.toolbar = toolbar


class SidePanel():

    """Incapsulates access to a side panel widget and its toolbar"""

    def __init__(self, widget):
        self.widget = widget
        self.views = {}       # view name (string) -> ViewAndToolbar instance


class IDEAccess():

    """Incapsulates access to the various IDE parts"""

    def __init__(self, parent):
        # The members below are initialized after
        # the 'activate' method is called
        self.settings = None
        self.globalData = None
        self.__sideBars = None
        self.__sidePanels = None
        self.__parent = parent

    def activate(self, ideSettings, ideGlobalData):
        """Saves references to the IDE settings and global data"""
        self.settings = ideSettings
        self.globalData = ideGlobalData

    def deactivate(self):
        """Resets the references to the IDE settings and global data"""
        self.settings = None
        self.globalData = None

    def showStatusBarMessage(self, message, timeout=10000):
        """Shows a temporary status bar message (default for 10sec)"""
        self.mainWindow.showStatusBarMessage(message, timeout)

    def clearStatusBarMessage(self):
        """Clears a temporary status bar message"""
        self.mainWindow.clearStatusBarMessage()

    @property
    def application(self):
        """Reference to the codimension application.

        See details in src/ui/application.py
        """
        if self.globalData is None:
            raise Exception("Plugin is not active")
        return self.globalData.application

    @property
    def mainWindow(self):
        """Reference to the application main window.
           See details in src/ui/mainwindow.py"""
        if self.globalData is None:
            raise Exception("Plugin is not active")
        return self.globalData.mainWindow

    @property
    def skin(self):
        """Reference to the current skin.

        See details in src/utils/skin.py
        """
        if self.globalData is None:
            raise Exception("Plugin is not active")
        return self.globalData.skin

    @property
    def project(self):
        """Reference to the current project.

        See details in src/utils/project.py
        Note: an object is provided even if there is no project loaded.
              To check if a project is loaded use
              getProject().isLoaded()
        """
        if self.globalData is None:
            raise Exception("Plugin is not active")
        return self.globalData.project

    @property
    def settingsDir(self):
        """The directory where the IDE setting files are stored.

        The directory is individual for each user and it is usually
        ~/.codimension3/
        """
        return SETTINGS_DIR

    @property
    def projectSettingsDir(self):
        """The directory where settings specific for the current
           project are stored. If there is no project loaded at the time of
           calling then None is returned.
           The directory is individual for each user/project and it is usually
           ~/.codimension3/<project UUID>"""
        if self.project.isLoaded():
            return self.project.userProjectDir
        return None

    @property
    def mainMenu(self):
        """Reference to the codimension main menu bar (QMenuBar)"""
        return self.mainWindow.menuBar()

    @property
    def statusBar(self):
        """Reference to the codimension main window status bar (QStatusBar)"""
        return self.mainWindow.statusBar()

    @property
    def mainToolbar(self):
        """Reference to the main window toolbar (QToolBar)"""
        return self.mainWindow.getToolbar()

    @property
    def editorsManager(self):
        """Reference to the editors manager (it derives QTabWidget)

        See details in src/ui/editorsmanager.py
        """
        return self.mainWindow.editorsManagerWidget.editorsManager

    @property
    def currentEditorWidget(self):
        """Reference to the current main area widget.

        Note: the widget could be of various types e.g. pixmap viewer,
              html viewer, text editor etc. All of them derive from
              MainWindowTabWidgetBase (see details in
              src/ui/mainwindowtabwidgetbase.py)
              The currentEditorWidget.getType() call provides the
              current widget type.
              The known widget types are described in
              src/ui/mainwindowtabwidgetbase.py
              Depending on the widget type various functionality is
              avalable. See the certain widget implementation files:
              PlainTextEditor     src/editor/texteditor.py
              PictureViewer       src/ui/pixmapwidget.py
              HTMLViewer          src/ui/htmltabwidget.py
              GeneratedDiagram    src/diagram/importsdgmgraphics.py
              ProfileViewer       src/profiling/profwidget.py
        """
        return self.editorsManager.currentWidget()

    @property
    def sidePanels(self):
        """Reference to a side panel widget map.

        I.e. a plugin class can use e.g. the following code:
        self.sidePanels[ "project" ].widget
        A side panel is identified by its string identifier.

        Supported panel names are (case sensitive):
        project, recent, classes, functions, globals, log,
        search, debuger, exceptions, breakpoints.

        Each side panel contains one or more views with thier toolbars. To
        get access to them a plugin class can use e.g. the following code:
        self.sidePanels[ "recent" ].views[ "files" ].widget
        self.sidePanels[ "recent" ].views[ "files" ].toolbar
        The names of views depend on the side panel.
        """
        if self.__sidePanels is None:
            self.__initializeSidePanels()
        return self.__sidePanels

    @property
    def sideBars(self):
        """Reference to a side bar widget map.

        I.e. a plugin class can use e.g. the following code:
        self.sideBars["bottom"].addTab(...)
        A bar is identified by the location. Supported locations:
        left, right, bottom, top
        A location can be None if there is no such a bar
        """
        if self.__sideBars is None:
            self.__sideBars = {'left': self.mainWindow._leftSideBar,
                               'right': self.mainWindow._rightSideBar,
                               'bottom': self.mainWindow._bottomSideBar,
                               'top': None}
        return self.__sideBars

    def __initializeSidePanels(self):
        """Initializes the side panels map"""
        self.__sidePanels = {}

        # Project
        projectPanel = SidePanel(self.mainWindow.projectViewer)
        projectPanel.views["project"] = ViewAndToolbar(
            projectPanel.widget.projectTreeView,
            projectPanel.widget.getProjectToolbar())
        projectPanel.views["fileSystem"] = ViewAndToolbar(
            projectPanel.widget.filesystemView,
            projectPanel.widget.getFileSystemToolbar())
        self.__sidePanels["project"] = projectPanel

        # Recent
        recentPanel = SidePanel(self.mainWindow.recentProjectsViewer)
        recentPanel.views["files"] = ViewAndToolbar(
            recentPanel.widget.recentFilesView,
            recentPanel.widget.getRecentFilesToolbar())
        recentPanel.views["projects"] = ViewAndToolbar(
            recentPanel.widget.projectsView,
            recentPanel.widget.getRecentProjectsToolbar())
        self.__sidePanels["recent"] = recentPanel

        # Classes
        classesPanel = SidePanel(self.mainWindow.classesViewer)
        classesPanel.views["classes"] = ViewAndToolbar(
            classesPanel.widget.clViewer,
            classesPanel.widget.toolbar)
        self.__sidePanels["classes"] = classesPanel

        # Functions
        funcPanel = SidePanel(self.mainWindow.functionsViewer)
        funcPanel.views["functions"] = ViewAndToolbar(
            funcPanel.widget.funcViewer,
            funcPanel.widget.toolbar)
        self.__sidePanels["functions"] = funcPanel

        # Globals
        globPanel = SidePanel(self.mainWindow.globalsViewer)
        globPanel.views["globals"] = ViewAndToolbar(
            globPanel.widget.globalsViewer,
            globPanel.widget.toolbar)
        self.__sidePanels["globals"] = globPanel

        # Log
        logPanel = SidePanel(self.mainWindow.logViewer)
        logPanel.views["log"] = ViewAndToolbar(
            logPanel.widget.messages,
            logPanel.widget.toolbar)
        self.__sidePanels["log"] = logPanel

        # Search
        searchPanel = SidePanel(self.mainWindow.findInFilesViewer)
        searchPanel.views["search"] = ViewAndToolbar(
            searchPanel.widget.getResultsTree(),
            searchPanel.widget.toolbar)
        self.__sidePanels["search"] = searchPanel

        # Debugger
        dbgPanel = SidePanel(self.mainWindow.debuggerContext)
        dbgPanel.views["variables"] = ViewAndToolbar(
            dbgPanel.widget.variablesViewer,
            None)
        dbgPanel.views["stack"] = ViewAndToolbar(
            dbgPanel.widget.stackViewer,
            None)
        dbgPanel.views["threads"] = ViewAndToolbar(
            dbgPanel.widget.threadsViewer,
            None)
        self.__sidePanels["debugger"] = dbgPanel

        # Exceptions
        excptPanel = SidePanel(self.mainWindow.debuggerExceptions)
        excptPanel.views["exceptions"] = ViewAndToolbar(
            excptPanel.widget.clientExcptViewer.exceptionsList,
            excptPanel.widget.clientExcptViewer.toolbar)
        excptPanel.views["ignoredExceptions"] = ViewAndToolbar(
            excptPanel.widget.ignoredExcptViewer.exceptionsList,
            excptPanel.widget.ignoredExcptViewer.toolbar)
        self.__sidePanels["exceptions"] = excptPanel

        # Breakpoints
        bpointPanel = SidePanel(self.mainWindow.debuggerBreakWatchPoints)
        bpointPanel.views["breakpoints"] = ViewAndToolbar(
            bpointPanel.widget.breakPointViewer.bpointsList,
            bpointPanel.widget.breakPointViewer.toolbar)
        self.__sidePanels["breakpoints"] = bpointPanel

    def sendLogMessage(self, level, msg, *args):
        """Sends a log message asynchronously.

        The method could be used safely from a non-GUI thread.

        level => integer, one of those found in logging:
                          logging.CRITICAL
                          logging.ERROR
                          logging.WARNING
                          logging.INFO
                          logging.DEBUG
        msg => message
        args => message arguments to be substituted (mgs % args)
        """
        try:
            self.__parent.pluginLogMessage.emit(level, msg % args)
        except Exception as exc:
            self.__parent.pluginLogMessage.emit(
                logging.ERROR,
                "Error sending a plugin log message. Error: " + str(exc))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Call trace viewer"""


import logging
from ui.qt import (QTreeWidget, QAbstractItemView, QWidget, QVBoxLayout, Qt,
                   QFrame, QAction, QToolBar, QSize, QSizePolicy, QRegExp,
                   QTreeWidgetItem, QApplication)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.spacers import ToolBarHSpacer, ToolBarExpandingSpacer
from ui.labels import HeaderFitLabel
from utils.settings import Settings
from utils.globals import GlobalData
from utils.pixmapcache import getIcon
from utils.project import CodimensionProject


class CallTraceBrowser(QTreeWidget):

    """Call trace browser implementation"""

    def __init__(self, parent=None):
        QTreeWidget.__init__(self, parent)

        self.setExpandsOnDoubleClick(False)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        self.setHeaderLabels(["", "From", "To"])

        self.__callStack = []
        self.__callIcon = getIcon('calltracecall.png')
        self.__retIcon = getIcon('calltracereturn.png')

        self.projectLoaded = False
        self.count = 0

        self.__entryRe = QRegExp(r"""(.+):(\d+)\s\((.*)\)""")
        self.__entryFormat = "{0}:{1} ({2})"

        self.itemActivated.connect(self.__itemActivated)

    def onProjectChanged(self):
        """Triggered when a project is changed"""
        self.projectLoaded = GlobalData().project.isLoaded()

    def clear(self):
        """Clears the data"""
        QTreeWidget.clear(self)
        self.__callStack = []
        self.count = 0

    def addCallTrace(self, isCall, fromFile, fromLine, fromFunction,
                     toFile, toLine, toFunction):
        """Adds a record to the list"""
        tooltip = 'Return\n'
        icon = self.__retIcon
        if isCall:
            tooltip = 'Call\n'
            icon = self.__callIcon

        if self.projectLoaded:
            project = GlobalData().project
            fromFile = project.getRelativePath(fromFile)
            toFile = project.getRelativePath(toFile)

        parentItem = self
        if self.__callStack:
            parentItem = self.__callStack[-1]

        fromItem = self.__entryFormat.format(fromFile, fromLine, fromFunction)
        toItem = self.__entryFormat.format(toFile, toLine, toFunction)
        item = QTreeWidgetItem(parentItem, ['', fromItem, toItem])
        item.setIcon(0, icon)
        item.setData(0, Qt.UserRole, isCall)
        item.setExpanded(True)

        tooltip += 'From: ' + fromItem + '\nTo: ' + toItem
        item.setToolTip(1, tooltip)
        item.setToolTip(2, tooltip)

        if isCall:
            self.__callStack.append(item)
        else:
            if self.__callStack:
                self.__callStack.pop(-1)

        self.count += 1

    def __itemActivated(self, item, column):
        """Double click on the item"""
        if item is not None and column > 0:
            columnStr = item.text(column)
            if self.__entryRe.exactMatch(columnStr.strip()):
                filename, lineno, _ = self.__entryRe.capturedTexts()[1:]
                try:
                    lineno = int(lineno)
                except ValueError:
                    return

            if self.projectLoaded:
                filename = GlobalData().project.getAbsolutePath(filename)

            editorsManager = GlobalData().mainWindow.editorsManager()
            editorsManager.openFile(filename, lineno)
            editor = editorsManager.currentWidget().getEditor()
            editor.gotoLine(lineno)
            editorsManager.currentWidget().setFocus()


class CallTraceViewer(QWidget):

    """Implements the call trace viewer"""

    def __init__(self, debugger, parent=None):
        QWidget.__init__(self, parent)

        self.__debugger = debugger
        self.__createLayout()

        self.__debugger.sigClientCallTrace.connect(self.__onCallTrace)
        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.clear()
            self.calltraceList.onProjectChanged()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__calltraceLabel = HeaderFitLabel(self)
        self.__calltraceLabel.setText('Call Trace')
        self.__calltraceLabel.setSizePolicy(QSizePolicy.Expanding,
                                            QSizePolicy.Fixed)
        self.__calltraceLabel.setMinimumWidth(10)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(18, 18))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__calltraceLabel)

        self.calltraceList = CallTraceBrowser(self)

        self.__startButton = QAction(
            getIcon('calltracestart.png'), "Start call tracing", self)
        self.__startButton.triggered.connect(self.__onStart)
        self.__startButton.setEnabled(not Settings()['calltrace'])

        self.__stopButton = QAction(
            getIcon('calltracestop.png'), "Stop call tracing", self)
        self.__stopButton.triggered.connect(self.__onStop)
        self.__stopButton.setEnabled(Settings()['calltrace'])

        self.__resizeButton = QAction(
            getIcon('resizecolumns.png'),
            "Resize the columns to their contents", self)
        self.__resizeButton.triggered.connect(self.__onResize)
        self.__resizeButton.setEnabled(True)

        self.__clearButton = QAction(
            getIcon('trash.png'), "Clear", self)
        self.__clearButton.triggered.connect(self.__onClear)
        self.__clearButton.setEnabled(False)

        self.__copyButton = QAction(
            getIcon('copymenu.png'), "Copy to clipboard", self)
        self.__copyButton.triggered.connect(self.__onCopy)
        self.__copyButton.setEnabled(False)

        # Toolbar
        self.toolbar = QToolBar()
        self.toolbar.setOrientation(Qt.Horizontal)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedHeight(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.__startButton)
        self.toolbar.addAction(self.__stopButton)

        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 15))
        self.toolbar.addAction(self.__resizeButton)
        self.toolbar.addAction(self.__copyButton)
        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 5))
        self.toolbar.addWidget(ToolBarExpandingSpacer(self.toolbar))
        self.toolbar.addAction(self.__clearButton)

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.toolbar)
        verticalLayout.addWidget(self.calltraceList)

    def __onCallTrace(self, isCall, fromFile, fromLine, fromFunction,
                      toFile, toLine, toFunction):
        """Call trace message received"""
        self.calltraceList.addCallTrace(isCall, fromFile, fromLine,
                                        fromFunction, toFile, toLine,
                                        toFunction)
        self.__clearButton.setEnabled(True)
        self.__copyButton.setEnabled(True)
        self.__updateHeader()

    def __updateHeader(self):
        """Updates the header"""
        count = self.calltraceList.count
        if count:
            self.__calltraceLabel.setText('Call Trace (' + str(count) + ')')
        else:
            self.__calltraceLabel.setText('Call Trace')

    def __onStart(self):
        """Start collecting calltrace"""
        self.__startButton.setEnabled(False)
        self.__stopButton.setEnabled(True)
        Settings()['calltrace'] = True
        self.__debugger.startCalltrace()

    def __onStop(self):
        """Stop collecting calltrace"""
        self.__startButton.setEnabled(True)
        self.__stopButton.setEnabled(False)
        Settings()['calltrace'] = False
        self.__debugger.stopCalltrace()

    def __onResize(self):
        """Resize the columns to its width"""
        for column in range(self.calltraceList.columnCount()):
            self.calltraceList.resizeColumnToContents(column)

    def __onClear(self):
        """Clears the view"""
        self.calltraceList.clear()
        self.__clearButton.setEnabled(False)
        self.__copyButton.setEnabled(False)
        self.__updateHeader()

    def clear(self):
        """Clears the view"""
        self.__onClear()

    def __onCopy(self):
        """Copy the content as text to clipboard"""
        content = []
        lhsMaxLength = 0
        try:
            item = self.calltraceList.topLevelItem(0)
            while item is not None:
                call = '<-'
                if item.data(0, Qt.UserRole):
                    call = '->'
                lhs = item.text(1)
                lhsLength = len(lhs)
                lhsMaxLength = max(lhsMaxLength, lhsLength)

                content.append([lhs, lhsLength, call + ' ' + item.text(2)])
                item = self.calltraceList.itemBelow(item)

            for index in range(len(content)):
                content[index] = \
                    content[index][0] + \
                    ' ' * (lhsMaxLength - content[index][1] + 1) + \
                    content[index][2]

            QApplication.clipboard().setText('\n'.join(content))
        except Exception as exc:
            logging.error('Error copying the call trace to clipboard: ' +
                          str(exc))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2018 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Provides the storage for the flow UI collapsed groups"""

import os.path
from .fileutils import loadJSON, saveJSON


class FlowUICollapsedGroups:

    """Loads/stores/saves the collapsed group list"""

    def __init__(self):
        # file name -> list of group ids
        self.__groups = {}
        self.__groupsFileName = None

    def reset(self):
        """Un-binds from the file system"""
        self.__groups = {}
        self.__groupsFileName = None

    def setup(self, dirName):
        """Binds the parameters to a disk file"""
        # Just in case - flush the previous data if they were bound
        FlowUICollapsedGroups.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for collapsed '
                            'groups. The given ' + dirName + ' is not.')

        self.__groupsFileName = dirName + 'collapsedgroups.json'
        if os.path.exists(self.__groupsFileName):
            FlowUICollapsedGroups.load(self)

    def load(self):
        """Loads the saved collapsed groups file"""
        if self.__groupsFileName:
            self.__groups = loadJSON(self.__groupsFileName,
                                     'collapsed groups', {})

    def save(self):
        """Saves the collapsed groups into a file"""
        if self.__groupsFileName:
            saveJSON(self.__groupsFileName, self.__groups, 'collapsed groups')

    def getFileGroups(self, fileName):
        """Provides None if not found"""
        return self.__groups.get(fileName, None)

    def setFileGroups(self, fileName, groups):
        """Sets the collapsed groups for the file"""
        if groups:
            self.__groups[fileName] = groups
            FlowUICollapsedGroups.save(self)
        else:
            if self.__groups.pop(fileName, None) is not None:
                FlowUICollapsedGroups.save(self)

    def addFileGroup(self, fileName, group):
        """Adds to a file group"""
        groups = self.__groups.get(fileName, None)
        if groups:
            if group not in groups:
                self.__groups[fileName].append(group)
                FlowUICollapsedGroups.save(self)
        else:
            self.__groups[fileName] = [group]
            FlowUICollapsedGroups.save(self)

    def removeFileGroup(self, fileName, group):
        """Removes from a file group"""
        groups = self.__groups.get(fileName, None)
        if groups:
            if group in groups:
                groups.remove(group)
                if groups:
                    self.__groups[fileName] = groups
                else:
                    self.__groups.pop(fileName, None)
                FlowUICollapsedGroups.save(self)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2019  Sergey V. Satskiy sergey.satskiy@tuta.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Resource cache for the markdown renderer"""

# pylint: disable=W0702
# pylint: disable=W0703

# By default the text browser does not download the internet resource so
# this facility needs to be implemented separately

# cml 1 gb id=0 title="stdlib imports"
import os.path
import logging
import datetime
import hashlib
import urllib.request
# cml 1 ge id=0
# cml 1 gb id=1 title="project imports"
from ui.qt import QThread, pyqtSignal, QObject
from .fileutils import loadJSON, saveJSON, saveBinaryToFile
# cml 1 ge id=1

TIMEOUT = 5   # timeout in seconds to do a request
CACHE_FILE_NAME = 'cachemap.json'


class ResourceRetriever(QThread):

    """Retrieves the item from the web"""

    sigRetrieveOK = pyqtSignal(str, str, str)       # url, uuid, file
    sigRetrieveError = pyqtSignal(str, str)         # url, file

    def __init__(self, parent=None):
        QThread.__init__(self, parent)
        self.__url = None
        self.__uuid = None
        self.__fName = None

    def get(self, url, fName, uuid):
        """Initiate the resource request"""
        self.__url = url
        self.__uuid = uuid
        self.__fName = fName
        self.start()

    def run(self):
        """Run the retriever"""
        try:
            req = urllib.request.urlopen(self.__url, timeout=TIMEOUT)
            saveBinaryToFile(self.__fName, req.read())
            self.sigRetrieveOK.emit(self.__url, self.__uuid, self.__fName)
        except Exception as exc:
            logging.error('Cannot retrieve %s: %s', self.__url, str(exc))
            self.sigRetrieveError.emit(self.__url, self.__fName)



class WebResourceCache(QObject):

    """The web resources cache"""

    sigResourceSaved = pyqtSignal(str, str, str)    # url, uuid, file

    def __init__(self, cacheDir):
        QObject.__init__(self)

        self.__urlToFileName = {}
        self.__threads = {}
        self.__cacheDir = os.path.normpath(cacheDir) + os.path.sep

        if os.path.exists(self.__cacheDir):
            if os.path.isdir(self.__cacheDir):
                if os.access(self.__cacheDir, os.W_OK):
                    self.__loadCache()
                    self.__saveCache()
                else:
                    logging.error('The web resource cache directory (%s) '
                                  'does not have write permissions. There '
                                  'will be no web resource downloading',
                                  self.__cacheDir)
                    self.__cacheDir = None
            else:
                logging.error('The web resource cache directory path (%s) '
                              'exists and is not a directory. There will be '
                              'no web resource downloading', self.__cacheDir)
                self.__cacheDir = None
        else:
            # Try to create the dir
            try:
                os.mkdir(self.__cacheDir)
            except Exception as exc:
                logging.error('Error creating the web resource cache directory '
                              '%s: %s '
                              'There will be no web resource downloading',
                              self.__cacheDir, str(exc))
                self.__cacheDir = None

    def __loadCache(self):
        """Loads the cache from the disk files"""
        if self.__cacheDir is None:
            return

        # Remove too old files
        now = datetime.datetime.now()
        limit = now - datetime.timedelta(days=1)
        limit = limit.timestamp()
        for item in os.listdir(self.__cacheDir):
            if item == CACHE_FILE_NAME:
                continue
            if os.path.isfile(self.__cacheDir + item):
                modtime = os.path.getmtime(self.__cacheDir + item)
                if modtime < limit:
                    try:
                        os.unlink(self.__cacheDir + item)
                    except Exception as exc:
                        logging.error('Error removing obsolete web resource '
                                      'file (%s): %s',
                                      self.__cacheDir + item, str(exc))

        if os.path.exists(self.__cacheDir + CACHE_FILE_NAME):
            prevCache = loadJSON(self.__cacheDir + CACHE_FILE_NAME,
                                 'web resources cache map', None)
            for item in prevCache.items():
                if os.path.exists(item[1]):
                    self.__urlToFileName[item[0]] = item[1]

    def __saveCache(self):
        """Saves the cache to the disk"""
        if self.__cacheDir is None:
            return
        dictToSave = {}
        for item in self.__urlToFileName.items():
            if item[1] is not None:
                dictToSave[item[0]] = item[1]
        saveJSON(self.__cacheDir + CACHE_FILE_NAME, dictToSave,
                 'web resources cache map')

    def onResourceSaved(self, url, uuid, fName):
        """Resource downloaded and saved"""
        self.__urlToFileName[url] = fName
        self.__onThreadFinish(fName)
        self.__saveCache()
        self.sigResourceSaved.emit(url, uuid, fName)

    def onResourceError(self, url, fName):
        """Error downloading the resource"""
        self.__urlToFileName[url] = None
        self.__onThreadFinish(fName)

    def __onThreadFinish(self, fName):
        """Cleans up after a retrieval thread"""
        thread = self.__threads.get(fName, None)
        if thread is not None:
            self.__disconnectThread(thread)
            thread.wait()
            self.__threads.pop(fName)

    def __connectThread(self, thread):
        """Connects the thread signals"""
        thread.sigRetrieveOK.connect(self.onResourceSaved)
        thread.sigRetrieveError.connect(self.onResourceError)

    def __disconnectThread(self, thread):
        """Connects the thread signals"""
        try:
            thread.sigRetrieveOK.disconnect(self.onResourceSaved)
            thread.sigRetrieveError.disconnect(self.onResourceError)
        except:
            pass

    def getResource(self, url, uuid):
        """Provides the resource"""
        if self.__cacheDir is None:
            return None

        if url in self.__urlToFileName:
            return self.__urlToFileName[url]

        fName = self.__cacheDir + hashlib.md5(url.encode('utf-8')).hexdigest()
        if fName in self.__threads:
            # Reject double request
            return None

        thread = ResourceRetriever()
        self.__threads[fName] = thread
        self.__connectThread(thread)
        thread.get(url, fName, uuid)
        return None



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Functions browser with hierarchy browsing capabilities"""

from utils.pixmapcache import getIcon
from .functionsbrowsermodel import FunctionsBrowserModel
from .objectsbrowserbase import ObjectsBrowser


class FunctionsBrowser(ObjectsBrowser):

    """Functions browser"""

    def __init__(self, parent=None):
        ObjectsBrowser.__init__(self, FunctionsBrowserModel(), parent)

        self.setWindowTitle('Functions browser')
        self.setWindowIcon(getIcon('icon.png'))



#!/usr/bin/python
# encoding: utf-8
# cml 1 cc bg=#ffffff
# File top comment

# cml 1 cc fg=#0000ff bg=#ffffff border=#ff0000
"""
File docstring
"""


#
# Code blocks
#


# cml 1 rt text="one"
# cml 1 gb id=1 title="Outer group"
# cml 1 gb id=0 title="A group"
alone = 1

one = 1
two = 2
# cml 1 ge id=0

# Leading
a = 10      # Side
b = 20      # Side
            # Side last


c = """
    ...
    """ + \
    "123456"    # Side
                # Another side
# cml 1 ge id=1

'''
...
'''


# cml 1 gb id=3 title="Empty group"
# cml 1 ge id=3



# cml 1 rt text="Quite a \"long\" text\nOn two lines"
x = 720


d = 154
e = 154  # side
f = 154

# cml 1 cc fg=#000000
alone = 2

# cml 1 cc fg=#000000
# Leading
a = 10      # Side
            # Side


#
# Imports
#


import sys, \
       os.path

# Leading
import sys      # Side
                # Side last

# Leading
from sys import path        # Side
                            # Side last


# cml 1 cc bg=#000000 fg=#eaeaea border=#ffffff
import sys, \
       os.path

# cml 1 cc bg=#eaeaea fg=#000000
# Leading
from sys import path        # Side
                            # Side last


#
# Functions
#


# Leading
def f( x ):     # Side
                # Last side

    # cml 1 cc fg=#ff0000
    """ f doc
        f doc line 2
    """
    pass

def f1( x,
        y ):
    " f1 doc "
    pass

def f2( x, y, z ):
    pass


# Leading
def f3( x,      # Side 1
        y,      # Side 2
        z ):
    " f3 doc "
    pass

def f4( a = """
            ...
            """ ):
    pass


# cml 1 rt text="My own header"
def fReplaced( x ):
    def nested():
        pass


#
# Classes
#


# cml 1 cc bg=#dddddd fg=#ffffff
# Leading
class C:    # Side
            # Last side
    # cml 1 cc bg=#555555 fg=#ffffff
    """
    Class C doc
    Class C doc line 2
    """
    pass    # Last class statement side 1
            # Last class statement side 2

# cml 1 gb id=2 title="Class in a group"
class C1( B1,
          B2 ):
    " C1 doc "
    # Leading inside group
    pass
# cml 1 ge id=2

# Leading
class C3( B1,       # Side 1
          B2,       # Side 2
          B3 ):
    # cml 1 cc bg=#ffffff
    " C3 doc "
    pass    # Side inside the class


class C4( C5 ):
    def member1( self ):
        pass
    # cml 1 cc bg=#777777 
    # Leading 1
    # Leading 2
    def member2( self,      # Side 1
                 x, y ):        # Side 2
                                # Side 3
        def nestedMember():
            pass


#
# Decorators
#

@decor1
def d():
    pass

# Leading
@decor2     # Side 1
            # Side 2
def d():
    pass

# Leading
@decor2( x,     # Side 1
         y )    # Side 2
def d():
    pass

@decor3
@decor4
class Cd:
    pass


#
# For
#


for x in y:
    pass

# Leading
for x in y:     # Side 1
    pass

for x in y:     # Side 1
                # Side 2
    pass

for x in y:
    pass
else:
    pass

for x in y:
    pass
# Leading
else:   # Side 1
        # Side 2
    pass

# Leading 1
# Leading 2
for x in y: # Side 1
            # Side 2
    pass
# Leading
else:   # Side 1
        # Side 2
    pass

#
# While
#

while True:
    pass

# Leading
while True:     # Side 1
    pass

while True:     # Side 1
                # Side 2
    pass

while True:
    pass
else:
    pass

while True:
    pass
# Leading
else:   # Side 1
        # Side 2
    pass

# Leading 1
# Leading 2ggggggggggggggggggg
while True: # Side 1
            # Side 2
    pass
# Leading
else:   # Side 1
        # Side 2
    pass



#
# Try
#

# cml 1 gb id=4 title=fff
try:
    pass
except:
    pass
# cml 1 ge id=4

if True:
    try:
        pass
    except Exception:
        pass
    except Exception as ex:
        pass
    except:
        pass
else:
    try:
        pass
    except Exception:
        pass


# Leading 1
try:        # Side 1
    pass
# Leading 2
# Leading 2.1gggggggggggggggggg
except Exception:   # Side 2
                    # Side 2.1
    pass
except Exception as ex:
    pass
# cml 1 doc anchor=000 title="Some anchor"
# Leading 4
except:         # Side 4
    pass

# Leading 1
try:
    a = 10
    b = 20  # Side 1
            # Side 2
# Leading 2
# Leading 2.1
except Exception:   # Side 2
                    # Side 2.1
    pass
except Exception as ex:
    pass
# Leading 4
except:         # Side 4
    pass
# Leading 5
# Leadin 5.1
else:   # Side 5
    pass
# Leading 6
finally:    # Side 6
    pass


#
# break/continue
#

for x in y:
    break

for x in y:
    # cml 1 cc bg=0,230,0 fg=255,255,255
    # Leading 1
    # Leading 2
    break   # Side 1
            # Side 2

while True:
    continue

# cml 1 cc bg=#dddddd fg=#343399
while True:
    # cml 1 cc bg=0,230,0 fg=255,255,255
    # Leading 1
    # Leading 2
    continue    # Side 1
                # Side 2


#
# Return
#

def f10():
    return

def f11():
    return None

def f12():
    return 154

def f13():
    return 154, \
           "Shakespeare"

def f14():
    # cml 1 cc bg="230,0,0" fg="255,255,255" border=#000000
    # Leading
    return 154  # Side 1
                # Side 2

def f15():
    # cml 1 sw
    # Leading
    return ( 154,               # Side 1
             "Shakespeare" )    # Side 2

# cml 1 cc bg=#454545 fg=#ffffff border=#ff0000
def f16():
    return """
           ...
           """

#
# Assert
#


assert x != 154

assert x != 154 and \
       y != 154 and \
       z != 154

# cml 1 cc bg=0,0,210 fg=255,255,255
# Leading
assert x != 154, ( "..."        # Side 1
                   "..." )      # side 2
                                # Side 3

assert """ one """, \
       """
       two
       """

#
# Raise
#

raise

raise Exception( "..." )

raise Exception( 1 +
                 2 +
                 3 )

# cml 1 cc bg=0,230,0 fg=0,0,230
# Leading
raise Exception( 1 +    # Side 1
                 3 )    # Side 2
                        # Side 3

raise """
      ...
      """

#
# sys.exit()
#

sys.exit( 0 )

sys.exit( 0 +
          0 +
          0 )

# Leading
sys.exit( 0 +   # Side 1
          0 )   # Side 2
                # Side 3

from sys import exit

exit( 0 )

exit( 0 +
      0 +
      0 )

# cml 1 cc bg="25, 200, 200"
# cml+     fg=#ffffff
# Leading
exit( 0 +   # Side 1
      0 )   # Side 2
            # Side 3

from sys import os, exit as EXIT

EXIT( 0 )

EXIT( 0 +
      0 +
      0 )

# Leading
EXIT( 0 +   # Side 1
      0 )   # Side 2
            # Side 3


#
# With
#

with open( "my.txt" ) as f:
    pass


with \
    open( "my.txt" ) \
        as \
            f:
    pass

# cml 1 cc bg=#dddddd fg=#343399 border=#ffffff
# leading
with open( "my.txt" +           # Side 1
           "your.txt" ) as f:   # Side 2
                                # Side 3
    pass



#
# If
#

if True:
    pass

# cml 1 sw
if True:
    pass
else:
    pass

# cml 1 rt text = "Some text"
if a == 154 and \
   b == 155 and \
   c == 156:
    pass
elif False:
    pass

    pass
else:
    pass


# Leading
# cml 1 rt text = ""
if True:    # Side 1
            # Side 2
    pass

    pass

    pass
# Leading 3
# cml 1 sw
# cml 1 cc bg=#345678 fg=#ffffff border=#ffffff
elif False: # Side 3
    pass
# Leading 4
else:       # Side 4
    def f( a,
           b,       # side
           c ) :
        " docstring "
        pass


# cml 1 cc bg=#ee00ee fg=#ffffff
# cml 1 rt text="Salary is too high"
if a > 456 and \
   b < 76 or \
   d == 99:
    pass
elif a < 456 and \
     b > 76 or \
     c == 99:
    pass
# else comment
else:
    pass

# cml 1 doc link=888 title=zdkslsdlks
def ff(x):
    if True:
        # Something

        if x > 32:
            x = 32

    # cml 1 doc anchor=some title=Yahoo bg=#6f6
    # cml+ border=#1010ff
    return """ddd %d
""" % 123

if '''fkfkf  %d
'''.replace(1,1) is None:
    pass
